(** printing ≃ %\ensuremath{\simeq}% *)
(** printing τ %\ensuremath{\tau}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Import tools atoms atoms_ops equiv_lst.

Parameter Name : Set.
Parameter τ_nm : Name -> list Atom.
Parameter beq_name : Name -> Name -> bool.
Axiom beq_name_spec : forall x y, reflect (x=y) (beq_name x y).
Axiom τ_nm_spec : forall a, NoDup (τ_nm a).

Definition is_perm x := exists p, p ⧓ (snd x) = τ_nm (fst x).
Parameter get_new : list Atom -> Atom.
Axiom new_atom : forall l, ~In (get_new l) l.

Fixpoint transfer_perm l m :=
  match (l,m) with
  | ([],_) |(_,[]) => []
  | (a::l,b::m) =>
    let q :=(transfer_perm l m) in
    let n := get_new (l++(elements q)) in
    ([b,n]++q++[a,n])
  end.
Lemma is_perm_no_dup_length m l :
  NoDup l -> NoDup m -> length l = length m ->
  (transfer_perm l m) ⧓ l = m.
Proof.
  revert m;induction l;intros;unfold is_perm in *.
  - simpl in *;destruct m;simpl in *;firstorder discriminate.
  - apply NoDup_cons_iff in H as (Il&H).
    destruct m.
    * discriminate.
    * inversion H1.
      apply NoDup_cons_iff in H0 as (Im&H0).
      set (q:=(transfer_perm l m));
        set (n:= get_new (l++(elements q))).
      assert (I: ~In n (l++(elements q))) by apply new_atom.
      rewrite In_app_iff in I.
      assert (HI: q⧓l=m) by (now apply IHl).
      replace (transfer_perm (_::_) (_::_)) with ([a0,n]++q++[a,n]);
        [|simpl;auto].
      repeat rewrite<- mact_comp_app;simpl.
      f_equal.
      -- simpl_beq.
         rewrite elements_inv_act by (now intro;apply I;right).
         now simpl_beq;case_eq_at n a0;simpl_beq.
      -- rewrite (@elements_inv_mact l).
         rewrite HI;apply elements_inv_mact.
         intros ? ? [->|[->|?]];simpl in *;apply I;auto.
         now apply elements_in_mact;rewrite HI.
         intros;simpl;intros [->|[->|h]];auto.
Qed.
Definition nodup_bool (l : list Atom):=
  (forallb (fun x => (count_occ Atom_eq_dec l x) =? 1) l).
Lemma nodup_bool_spec l : reflect (NoDup l) (nodup_bool l).
Proof.
  apply iff_reflect.
  unfold nodup_bool;rewrite forallb_forall,
                    (NoDup_count_occ' Atom_eq_dec).
  setoid_rewrite Nat.eqb_eq;firstorder.
Qed.

Definition is_perm_bool x :=
  (length (snd x) =? length (τ_nm (fst x))) && (nodup_bool (snd x)).

Lemma is_perm_bool_spec x : reflect (is_perm x) (is_perm_bool x).
Proof.
  apply iff_reflect. unfold is_perm_bool.
  rewrite andb_true_iff, Nat.eqb_eq,
  <-(reflect_iff _ _ (nodup_bool_spec _)).
  split;intro.
  - destruct H as (p&h).
    replace (snd x) with (p⋆⧓(τ_nm (fst x)))
      by (now rewrite<-h;simpl_perm); split.
    -- now rewrite mact_length.
    -- apply NoDup_mact,τ_nm_spec.
  - exists (transfer_perm (snd x)(τ_nm (fst x))).
    apply is_perm_no_dup_length;intuition.
    apply τ_nm_spec.
Qed.
Lemma is_perm_bool_is_perm x :
   is_perm x <-> is_perm_bool x = true.
Proof. apply (reflect_iff _ _ (is_perm_bool_spec _)). Qed.
  
Definition Letter := { x | is_perm_bool x = true}.
Definition beq_letter : Letter -> Letter -> bool :=
  fun n1 n2 =>
    match (proj1_sig n1,proj1_sig n2) with
    |((x,l),(y,m)) =>
     if (beq_name x y)
     then beq_list l m
     else false
    end.
Definition eq_letter (x y : Letter) := proj1_sig x = proj1_sig y.
Infix " ≃ " := eq_letter (at level 50).
Hint Unfold eq_letter.
  
(* Axiom Letter_proof_irr : forall x y : Letter, *)
(*     proj1_sig x = proj1_sig y -> x = y. *)
Global Instance eq_letter_Equivalence : Equivalence eq_letter.
Proof.
  split.
  * firstorder. 
  * firstorder. 
  * now intros (?,?) (?,?) (?,?) ? ?;unfold eq_letter in *;simpl in *;
    firstorder;transitivity x0.
Qed.
Lemma beq_letter_spec : forall x y, reflect (x ≃ y) (beq_letter x y).
Proof.
  intros ((x,l),a) ((y,m),b);unfold beq_letter;simpl;case beq_name_spec.
  - case beq_list_spec.
    * now intros -> ->; apply ReflectT.
    * now intros E _; apply ReflectF;intros h;inversion h.
  - now intros E; apply ReflectF;intros h;inversion h.
Qed.
Definition τ : Letter -> (list Atom) :=
  fun n =>
    match (proj1_sig n) with
    | (x,l) => l
    end.
Ltac simpl_is_perm_hyp p h :=
  apply is_perm_bool_is_perm in h as (p&h); simpl in h.
Ltac simpl_is_perm :=
  apply is_perm_bool_is_perm;simpl.
Tactic Notation "simpl_is_perm" "in" ident(h):=
  let p := fresh "p" in
   simpl_is_perm_hyp p h.
Tactic Notation "simpl_is_perm" "in" ident(h) "as" ident(p) :=
   simpl_is_perm_hyp p h.
Lemma τ_spec : forall a, NoDup (τ a).
Proof.
  intros ((x,l),h);unfold τ;simpl in *; simpl_is_perm in h as p.
  rewrite <- (mact_inv_l p),h,<-NoDup_mact;apply τ_nm_spec.
Qed.
Lemma eq_letter_τ : Proper (eq_letter ==> Logic.eq) τ.
Proof. intros ((?,?),?) ((?,?),?) h;unfold τ,eq_letter in *;
       inversion h;auto.
Qed.
