(** printing π %\ensuremath{\pi}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Import proofs perm theories order.
(** * Strange stuff *)

Fixpoint can_var (n : nat) : list Atom :=
  match n with
  | O => []
  | S n => get_new (can_var n)::(can_var n)
  end.
Lemma can_var_NoDup: forall n, NoDup (can_var n).
Proof.
  induction n;simpl;auto using NoDup_nil.
  apply NoDup_cons;auto;apply new_atom.
Qed.
Lemma can_var_length : forall n, length (can_var n) = n.
Proof. induction n;simpl;auto. Qed.

Definition conv_lst_perm : list Atom -> perm :=
  fun l => transfer_perm (can_var (length l)) l.
Lemma conv_lst : forall l, NoDup l ->
                           conv_lst_perm l ⧓ (can_var (length l)) = l.
Proof.
  intros;unfold conv_lst_perm;apply is_perm_no_dup_length;
  auto using can_var_NoDup,can_var_length.
Qed.
Lemma conv_can_var : forall n, conv_lst_perm (can_var n) ≌ nil.
Proof.
  unfold conv_lst_perm;intro n.
  rewrite can_var_length;induction (can_var n);
  simpl;auto.
  intros ?;simpl.
  pose proof (@new_atom (l++ (elements (transfer_perm l l)))) as h.
  revert h; set (x:=get_new (l++ (elements (transfer_perm l l)))).
  simpl_in;intro h.
  rewrite <-act_comp_app,IHl;simpl.
  case_eq_at (a0) x;simpl_beq.
  case_eq_at x a;repeat simpl_beq;auto.
  case_eq_at x a;repeat simpl_beq;auto.
  case_eq_at a0 a;repeat simpl_beq;auto.
Qed.

Lemma conv_lst_inv l :
  NoDup l -> (conv_lst_perm l)⋆ ⧓ l = (can_var (length l)).
Proof. now intro h;rewrite <- (conv_lst h) at 2;simpl_perm. Qed.
  
Inductive eq_perm_fresh : relation expr :=
| eq_θ_fresh p q e : (forall a, p⋈a <> q⋈a -> a # e) ->
                     eq_perm_fresh (‹ p ›e) (‹ q ›e)
| eq_θ_fr_nil e : eq_perm_fresh (‹[]›e) e
| eq_θ_fr_id : forall (A B : list Atom) (p : perm),
    B ≈ p ⧓ A -> eq_perm_fresh (‹ p › id A) (id B).

Fixpoint convert e :=
  match e with
  (* | 0 => 0 *)
  (* | 1 => 1 *)
  (* | ⊥ A => ‹conv_lst_perm 〈A〉› ⊥ (can_var (length 〈A〉)) *)
  (* | id A => ‹conv_lst_perm 〈A〉› id (can_var (length 〈A〉)) *)
  | vat a => ‹conv_lst_perm [a]› ((conv_lst_perm [a])⋆ • vat a)
  | var x => ‹conv_lst_perm (τ x)› ((conv_lst_perm (τ x))⋆ • var x)
  | ν a e => ν a (convert e)
  | ω a e => ω a (convert e)
  | e + f => convert e + convert f
  | e · f => convert e · convert f
  | e^* => convert e^*
  | ‹ p › e => ‹ p › convert e
  | _ => e
  end.
Lemma convert_involutive e :
  eq_perm_fresh ⊢ convert (convert e) ≡ (convert e).
Proof.
  induction e;simpl;auto.
  (* - rewrite <-(NoDup_undup_eq (can_var_NoDup _)). *)
  (*   rewrite conv_can_var, can_var_length. *)
  (*   apply eq_θ,eq_ax,eq_θ_fr_nil;auto. *)
  (* - rewrite <-(NoDup_undup_eq (can_var_NoDup _)). *)
  (*   rewrite conv_can_var, can_var_length. *)
  (*   apply eq_θ,eq_ax,eq_θ_fr_nil;auto. *)
  - apply eq_θ;auto;pose proof (τ_spec l) as h;
    destruct l as ((?,?),?);unfold τ in *;simpl in *.
    rewrite (conv_lst_inv h) at 1.
    rewrite conv_can_var.
    etransitivity;[apply eq_ax,eq_θ_fr_nil;auto|].
    apply eq_var;unfold eq_letter;simpl;f_equal.
    repeat rewrite (conv_lst_inv h).
    pose proof (equiv_perm_inverse(conv_can_var (length l))).
    now rewrite (equiv_perm_mact _ _ H);simpl;simpl_perm.
  - assert(h:NoDup [a]) by (now repeat solve_no_dup).
    (* pose proof (conv_lst_inv h) as h';simpl in h'. *)
    repeat simpl_beq.
    (* repeat rewrite h'. *)
    rewrite (equiv_perm_inverse(conv_can_var 1));simpl.
    (* rewrite (conv_can_var 1). *)
    apply eq_θ,eq_ax,eq_θ_fr_nil;auto.
Qed.
    
    
  
Lemma convert_eq_perm e :
  eq_perm_elim ∨ eq_perm ⊢ e ≡ convert e.
Proof.
  induction e;simpl;auto.
  (* *  eql;eql;symmetry; apply eq_ax, eq_θ_bot. *)
  (*    rewrite conv_lst;try solve_no_dup;intro;simpl_in;tauto. *)
  (* *  eql;eqr;symmetry; apply eq_ax, eq_θ_id. *)
  (*    rewrite conv_lst;try solve_no_dup;intro;simpl_in;tauto. *)
  * etransitivity;[|eql;eapply eq_θ,eq_ax,eq_θ_var;reflexivity].
    etransitivity;[|symmetry;eqr;eapply eq_ax,eq_θ_θ].
    rewrite perm_p_pinv_eq_nil;auto 10.
  * etransitivity;[|eql;symmetry;eapply eq_ax,eq_θ_at;reflexivity].
    unfold conv_lst_perm;simpl;repeat simpl_beq.
    case_eq_at (get_new []) a;simpl_beq;auto.
Qed.

Lemma convert_untyped e : untyped e <-> untyped (convert e).
Proof. induction e;simpl in *;firstorder. Qed.
Lemma convert_str_positive e :
  str_positive e <-> str_positive (convert e).
Proof. induction e;simpl in *;firstorder. Qed.
Lemma convert_single e : single e <-> single (convert e).
Proof. induction e;simpl in *;firstorder. Qed.
Lemma convert_multi e : multi e <-> multi (convert e).
Proof. induction e;simpl in *;firstorder. Qed.

Definition NKA_nice :=
  (ka_eq_u∨eq_perm∨eq_perm_fresh
   ∨eq_ν_smpl∨eq_nom_untyped_p∨eq_nom_untyped).

Lemma convert_fresh a e : a # e <-> a # convert e.
Proof. revert a;induction e;intros;simpl in *;try (now firstorder).
       now destruct l as ((?,?),?);unfold τ;simpl; simpl_in;simpl_perm.
       repeat simpl_beq.
       case_eq_at a0 a;repeat simpl_beq;try tauto.
       case_eq_at a0 (get_new []);repeat simpl_beq;firstorder.
Qed.

Lemma untyped_NKA_nice e f :
  NKA_nice ⊢ e ≡ f -> untyped e <-> untyped f.
Proof.
  intro E;induction E;simpl in *;try (now firstorder).
  destruct H;simpl in *;firstorder.
  destruct H;simpl in *;firstorder.
  destruct H as [[[[[E|E]|E]|E]|E]|E];destruct E;simpl in *;firstorder.
Qed.  
Lemma str_positive_NKA_nice e f :
  NKA_nice ⊢ e ≡ f -> str_positive e <-> str_positive f.
Proof.
  intro E;induction E;simpl in *;try (now firstorder).
  destruct H;simpl in *;firstorder.
  destruct H;simpl in *;firstorder.
  destruct H as [[[[[E|E]|E]|E]|E]|E];destruct E;simpl in *;firstorder.
Qed.


Lemma NKA_nice_stronger_NKA_θp e f : str_positive e -> untyped e ->
                       str_positive f -> untyped f ->
                       NKA_nice ⊢ e ≡ f -> NKA_θp ⊢ e ≡ f.
Proof.
  intros se ue sf uf E;induction E;simpl in *;intuition.
  * now rewrite H0,IHE2;auto;rewrite <-(str_positive_NKA_nice E1)
                             || rewrite<- (untyped_NKA_nice E1).
  * destruct H;simpl in *;firstorder;eauto.
  * unfold NKA_nice,NKA_θp in *;autounfold in H;firstorder;
    auto 10.
    destruct H;auto;simpl in *.
    revert p q H;induction e;intros;simpl in *;try tauto.
  - etransitivity;[|symmetry];eql;eql;eql;eqr;apply eq_ax,eq_θ_1.
  - assert (hp:p⧓τ l = q⧓τ l).
    unfold mact;apply map_ext_in;firstorder;
    case_eq_at (p⋈a) (q⋈a);firstorder.
    etransitivity;[|symmetry;etransitivity];
    try (eql;eql;eql;eqr;apply eq_ax,eq_θ_var).
    apply eq_var;destruct l as ((?,?),?);unfold τ,eq_letter in *;
    simpl in *;rewrite hp;auto.
  - assert (h:p⋈a=q⋈a).
    case_eq_at (p⋈a) (q⋈a);firstorder.
    apply H in h;firstorder.
    etransitivity;[|symmetry;etransitivity];
    try (eql;eql;eql;eqr;apply eq_ax,eq_θ_at);auto.
  - transitivity (‹ p ›e1+‹ p ›e2); auto 10.
    transitivity (‹ q ›e1+‹ q ›e2); auto 10.
    apply eq_plus;[apply IHe1|apply IHe2];firstorder.
  - transitivity (‹ p ›e1·‹ p ›e2); auto 10.
    transitivity (‹ q ›e1·‹ q ›e2); auto 10.
    apply eq_times;[apply IHe1|apply IHe2];firstorder.
  - transitivity ((‹ p ›e)^* ); auto 10.
    transitivity ((‹ q ›e)^* ); auto 10.
  -  transitivity (ν (p⋈a) (‹ p ›e)); auto 10.
     transitivity (ν (q⋈a) (‹ q ›e)); auto 10.
     assert (exists b, b#e/\p⋈b=b/\q⋈b=b) as (b&hbe&hpb&hqb).
     exists (get_new (support e++elements p++elements q)).
     clear se ue sf uf IHe H;intuition;
     apply fresh_support||apply elements_inv_act;
     intro;apply (@new_atom(support e ++ elements p ++ elements q));
     simpl_in;intuition.
     assert (([b,p⋈a]++p)⋈a = b).
     rewrite <-act_comp_app;simpl;simpl_beq.
     rewrite <-hpb at 1;case beq_atom_refl;rewrite perm_bij;
     [intros ->|];intuition.
     assert (([b,q⋈a]++q)⋈a = b).
     rewrite <-act_comp_app;simpl;simpl_beq.
     rewrite <-hqb at 1;case beq_atom_refl;rewrite perm_bij;
     [intros ->|];intuition.
     assert (forall x : Atom, ([b,p⋈a]++p) ⋈ x <> ([b,q⋈a]++q) ⋈ x
                              -> x # e).
     intro x;repeat rewrite <-act_comp_app;simpl.
     rewrite <- hpb at 1;rewrite <- hqb at 3.
     case beq_atom_refl;rewrite perm_bij.
     intros ->;simpl_beq;firstorder.
     intro;case (beq_atom_refl (q ⋈ x) (q ⋈ a));rewrite perm_bij.
     intros ->;simpl_beq.
     case beq_atom_refl;rewrite perm_bij;firstorder.
     case beq_atom_refl;rewrite perm_bij.
     intros ->;firstorder.
     intros h _;case beq_atom_refl;rewrite perm_bij;firstorder.
     transitivity (ν b (‹[b,p⋈a]++p›e)); auto 10.
     symmetry;etransitivity.
     symmetry;eql;eql;eql;eql;eqr;apply eq_ν,eq_ax,eq_θ_θ.
     symmetry; eql;eqr;apply eq_ax,equ_θν'.
     simpl;rewrite <-hpb;simpl_perm;auto.
     symmetry;transitivity (ν b (‹[b,q⋈a]++q›e));auto 10.
     symmetry;etransitivity.
     symmetry;eql;eql;eql;eql;eqr;apply eq_ν,eq_ax,eq_θ_θ.
     symmetry;eql;eqr;apply eq_ax,equ_θν'.
     simpl;rewrite <-hqb;simpl_perm;auto.
  - etransitivity.
    eql;eql;eql;eql;eqr;apply eq_ax,eq_θ_θ.
    symmetry;etransitivity;[|symmetry].
    eql;eql;eql;eql;eqr;apply eq_ax,eq_θ_θ.
    apply IHe;auto.
    intro;repeat rewrite <-act_comp_app.
    now intro h;apply H in h;simpl_perm.
  - auto 10.
Qed.
Lemma equiv_lst_length_undup l m: l≈m -> length 〈l〉 = length 〈m〉.
Proof.
  revert m;induction l;simpl;intuition.
  destruct m;simpl;auto.
  exfalso; unfold equiv_lst in *;simpl in *;rewrite (H a);tauto.
  case inb_spec;intro; simpl.
  apply IHl;intro x;rewrite <- (H x);simpl_in;split;[|intros [->|h]];
  tauto.
  transitivity (length (a::〈rm a m〉)).
  simpl;f_equal; apply IHl;intro x;simpl_in;
  rewrite <- (H x);simpl_in;case_eq_at x a;intuition.
  assert (h:In a 〈m〉) by (simpl_in;rewrite <- (H a);now left). 
  rewrite undup_rm.
  pose proof (in_split _ _ h) as (l1 & l2 & hm);rewrite hm.
  assert (h1 : NoDup 〈m〉) by (solve_no_dup).
  rewrite hm in h1. apply NoDup_remove in h1.
  simpl;rewrite rm_app;repeat rewrite app_length;simpl.
  rewrite if_Atom_eq_dec_eq;auto.
  repeat rewrite rm_id;simpl_in;try tauto;lia.
Qed.
(* Lemma conv_lst_inv b : NoDup b -> *)
(*                        (conv_lst_perm b ⋆) ⧓ b = can_var (length b). *)
(* Proof. now intro;rewrite <-(@conv_lst b) at 2. Qed. *)
Lemma mact_eq_act_eq  p1 p2 l:
  p1 ⧓ l = p2 ⧓ l -> forall x,In x l -> p1 ⋈ x = p2 ⋈ x.
    induction l;simpl;firstorder.
    now subst;inversion H.
    inversion H;firstorder.
Qed.
Lemma eq_perm_elim_convert e f:
  eq_perm_elim e f -> NKA_nice ⊢ convert e ≡ convert f.
Proof.
  intro E;destruct E;simpl;auto.
  * eql;eql;eql;eqr;etransitivity;[|apply eq_ax,eq_θ_fr_nil].
    apply eq_ax,eq_θ_fresh;intuition.
  * etransitivity.
    eql;eql;eql;eql;eqr;apply eq_ax,eq_θ_θ.
    assert (NoDup [a]/\NoDup [p⋈a]) as (ha & hpa)
        by (now split;repeat solve_no_dup).
    repeat simpl_beq;unfold conv_lst_perm;simpl.
    eql;eql;eql;eqr;apply eq_ax,eq_θ_fresh;simpl;
    intros x h1 ->;apply h1;repeat rewrite <-act_comp_app.
    simpl;repeat simpl_beq;case_eq_at (get_new []) a;simpl_beq.
    case_eq_at (p⋈a) a;simpl_beq;auto.
    case_eq_at (p⋈a) (get_new []);simpl_beq;auto. 
  * etransitivity.
    eql;eql;eql;eql;eqr;apply eq_ax,eq_θ_θ.
    pose proof (τ_spec a) as nd.
    assert (NoDup (p⧓(τ a))) as ndp by (now solve_no_dup).
    apply conv_lst_inv in nd;apply conv_lst_inv in ndp;simpl in *.
    assert (hl:length (τ a) = length (p⧓τ a))
      by (symmetry;apply mact_length).
    destruct a as ((?,?),?);unfold τ in *;simpl in *.
    rewrite hl,<-ndp,mact_comp_app in nd.
    etransitivity;[|apply eq_θ;[reflexivity|]].
    eql;eql;eql;eqr;apply eq_ax,eq_θ_fresh;simpl;unfold τ;simpl.
    intro a;intros h1 h2;apply h1;simpl_in;simpl_perm.
    apply (mact_eq_act_eq nd) in h2;simpl_perm.
    now rewrite h2 at 2;repeat rewrite<-act_comp_app;simpl_perm.
    now apply eq_var;unfold eq_letter;simpl;
    rewrite mact_comp_app;f_equal.
  * now eql;eql;eql;eqr;apply eq_ax,eq_θ_fr_id.
  * eql;eql;eql;eqr;apply eq_ax,eq_θ_fr_nil.
Qed.
    
  
Lemma convert_eq e f : str_positive e -> untyped e ->
                       str_positive f -> untyped f ->
                       NKA_θp ⊢ e ≡ f <-> NKA_nice ⊢ convert e ≡ convert f.
Proof.
  split.
  - intro E; induction E;simpl in *;intuition.
    -- now rewrite H4;[apply IHE2| |];auto;
       rewrite <-(eq_str_positive NKA_θp_str_positive E1)
       || rewrite<- (eq_untyped NKA_θp_untyped E1).
    -- destruct x as ((x,tx),px); destruct y as ((y,ty),py);
       unfold eq_letter in H3;simpl in H3;inversion H3;unfold τ;
       simpl;auto; subst;apply eq_θ,eq_var;auto.
    -- destruct H3;simpl;auto.
    -- destruct H3;simpl in *;intuition;eauto 10.
    -- destruct H3 as [[[[[E|E]|E]|E]|E]|E];unfold NKA_nice;simpl;
       try (now destruct E;simpl;auto 10).
       --- now apply eq_perm_elim_convert.
       --- destruct E;simpl.
           now rewrite convert_fresh in H3; eql;eqr;
           apply eq_ax, equ_θν'.
       --- destruct E;simpl.
           now rewrite convert_fresh in H3; eqr; apply eq_ax, equ_νω1.
           now rewrite convert_fresh in H3; eqr; apply eq_ax, equ_ν3.
           now rewrite convert_fresh in H3; eqr; apply eq_ax, equ_ν4.
  - intros.
    etransitivity;[| symmetry;etransitivity];
    try (eapply stronger_ax_proof;
          [|apply convert_eq_perm;auto]);
    auto; unfold NKA_θp;autounfold;intuition;
    symmetry; apply NKA_nice_stronger_NKA_θp;auto;
    rewrite <- (convert_untyped) || rewrite <- (convert_str_positive);
    auto.
Qed.

Fixpoint nice e : Prop :=
  match e with
  | 0 | 1 | ⊥ _ | id _ => True
  | var l =>  τ l = can_var (length (τ l))
  | vat a => [a] = can_var 1
  | e + f | e · f => nice e /\ nice f
  | e^* | ν _ e | ω _ e | ‹ _ › e => nice e
  end.
Lemma nice_convert e : nice (convert e).
Proof.
  induction e;simpl;firstorder;try rewrite can_var_length;auto.
  - now pose proof (τ_spec l) as h;destruct l as ((?,?),?);
    unfold τ in *;simpl in *;rewrite (conv_lst_inv h),can_var_length.
  - transitivity (can_var (length [a])).
    now rewrite <- conv_lst_inv;simpl;repeat solve_no_dup.
    now simpl.
Qed.

Lemma nice_NKA_nice e f : NKA_nice e f -> nice e <-> nice f.
Proof.
  intro E;destruct E as [[[[[E|E]|E]|E]|E]|E];destruct E;simpl;
  intuition.
Qed.
Lemma nice_eq ax : Proper (ax ==> iff) nice ->
                   Proper (eq ax ==> iff) nice.
Proof.
  intros h e f E;induction E;simpl;try tauto.
  - now destruct x as ((?,?),?);destruct y as ((?,?),?);
    unfold τ,eq_letter in *;simpl in *;inversion H.
  - destruct H;simpl;tauto.
  - destruct H;simpl in *;tauto.
  - now apply h.
Qed.
Lemma nice_convert_eq e : nice e -> eq_perm_fresh ⊢ convert e ≡ e.
Proof.
  induction e;simpl;intuition.
  - rewrite H;
    pose proof (conv_can_var (length (τ l)));simpl.
    rewrite H0 at 1.
    pose proof (apply_perm_var_equiv_perm l (equiv_perm_inverse H0))
      as h;unfold Basics.flip in h.
    etransitivity;[apply eq_θ,eq_var;eauto|clear h H0].
    etransitivity.
    apply eq_ax,eq_θ_fr_nil.
    apply eq_var;destruct l as ((?,?),?);unfold eq_letter;simpl;
    now simpl_perm.
  - rewrite H.
    rewrite (conv_can_var 1) at 1;simpl.
    etransitivity.
    apply eq_ax,eq_θ_fr_nil.
    repeat simpl_beq;inversion H;reflexivity.
Qed.
Lemma convert_eq2 e f : str_positive e -> untyped e -> nice e -> 
                        str_positive f -> untyped f -> nice f ->
                       NKA_nice ⊢ convert e ≡ convert f <-> NKA_nice ⊢ e ≡ f.
Proof.
  intros se ue ne sf uf nf.
  apply nice_convert_eq in ne.
  apply nice_convert_eq in nf.
  assert (eq_perm_fresh ⇒ NKA_nice) as S.
  unfold stronger,NKA_nice;firstorder.
  apply (stronger_ax_proof S) in ne.
  apply (stronger_ax_proof S) in nf.
  rewrite ne,nf;tauto.
Qed.

Definition Nice := (nice ⊗ single ⊗ str_positive ⊗ untyped,NKA_nice).

Lemma convert_morph : convert ⊩ ((fst NKAspu)^2,snd NKAspu) →((fst Nice)^2,snd Nice).
Proof.
  unfold NKAspu,Nice;intros e f;split;simpl in *.
  - firstorder;rewrite <- convert_str_positive || rewrite <- convert_untyped
               || rewrite <- convert_single || apply nice_convert;auto.
  - apply convert_eq;firstorder.
Qed.
Lemma id_morph : (fun x => x) ⊩ ((fst Nice)^2,snd Nice) → ((fst NKAspu)^2,snd NKAspu) .
Proof.
  unfold NKAspu,Nice;intros e f;split;simpl in *.
  - firstorder.
  - rewrite <- convert_eq2,convert_eq; firstorder.
Qed.

Lemma NKAspu_Nice : NKAspu ⊲ Nice.
Proof. eexists; eapply convert_morph. Qed.
Lemma Nice_NKAspu : Nice ⊲ NKAspu.
Proof. eexists; eapply id_morph. Qed.