(** printing π %\ensuremath{\pi}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Import proofs.


Lemma apply_perm_equiv_perm p q e :
  p≌q -> FFalse ⊢ p•e ≡ q•e.
Proof.
  intro E; apply eq_expr_eq.
  eapply (apply_perm_equiv_perm E);reflexivity.
Qed.
Lemma eq_rm_perm : Proper (eq FFalse ==> eq FFalse) rm_perm.
Proof. intros e f E;induction E;simpl;auto.
       now rewrite IHE1.
       destruct H;simpl;auto.
       destruct H;simpl in *;eauto.
Qed.
Lemma eq_rm_perm_bis :
  forall e p, (eq_perm ∨ eq_perm_elim)
              ∨ zero_perm_u ⊢rm_perm (p•e) ≡ p•e.
Proof.
  induction e;intros q;simpl in *;auto;
  try tauto.
  transitivity ((q++p)•e);[|symmetry];auto.
Qed.

Lemma zero_perm_u_apply_perm :
  forall p, Proper (zero_perm_u ==> (eq FFalse))
                   (rm_perm ∘ (apply_perm p)).
Proof. intros p e f h;destruct h;auto. Qed.
Lemma zero_perm_t_apply_perm :
  forall p, Proper (zero_perm_t ==> (eq FFalse))
                   (rm_perm ∘ (apply_perm p)).
Proof. intros p e f h;destruct h;autounfold;now rewrite H,mact_comp_app. Qed.

Lemma eq_perm_apply_perm :
  forall p, Proper (eq_perm ==> (eq FFalse))
                   (rm_perm ∘ (apply_perm p)).
Proof. unfold Basics.compose;intros p e f h;destruct h;simpl;auto.
       now rewrite act_comp_app.
       now rewrite app_ass.
Qed.       
Lemma eq_perm_ω_apply_perm :
  forall p, Proper (eq_perm_ω ==> (eq FFalse))
                   (rm_perm ∘ (apply_perm p)).
Proof. unfold Basics.compose;intros p e f h;destruct h;simpl;auto.
       now rewrite act_comp_app.
Qed.
Lemma eq_perm_elim_apply_perm :
  forall p, Proper (eq_perm_elim ==> (eq FFalse))
                   (rm_perm ∘ (apply_perm p)).
Proof. unfold Basics.compose;intros p e f h;destruct h;simpl;auto.
       now rewrite act_comp_app.
       now apply eq_expr_eq;simpl;rewrite apply_perm_var_app.
       now rewrite H,mact_comp_app.
       now rewrite app_nil_r.
Qed.
    
Lemma eq_nom_untyped_p_apply_perm :
  forall p, Proper (eq_nom_untyped_p ==> (eq eq_nom_untyped_np))
                   (rm_perm ∘ (apply_perm p)).
Proof.
  unfold and_rel,rel;unfold Basics.compose;intros p e f h;
  destruct h;simpl in *; etransitivity.
  - apply eq_ax; apply(@equ_θν (p⋈a)).
    now apply fresh_rm_perm.
  - apply eq_expr_eq;simpl;intuition.
    rewrite <-(rm_perm_apply_perm_swap _ _ );apply eq_expr_rm_perm.
    apply morphisms.apply_perm_equiv_perm;[|reflexivity].
    clear e H;intro x;repeat rewrite<-act_comp_app;simpl.
    case_eq_at x a;repeat simpl_beq;auto.
    case_eq_at x b;repeat simpl_beq;auto.
    now rewrite<- (perm_bij p) in h; simpl_beq.
    now rewrite<- (perm_bij p) in *; repeat simpl_beq.
Qed.
Lemma eq_nom_typed_p_apply_perm :
  forall p, Proper (eq_nom_typed_p ==> (eq eq_nom_typed_np))
                   (rm_perm ∘ (apply_perm p)).
Proof.
  unfold and_rel,rel in *;unfold Basics.compose;intros p e f h;
  destruct h;simpl in *; destruct H as (I & H);
  etransitivity.
  - apply eq_ax;eapply(@eqt_θν _ (p⋈a)).
    simpl;split;[rewrite In_mact,inverse_comp_l;apply I|].
    apply has_type_rm_perm_apply_perm;auto.
    now apply has_type_apply_perm;simpl;simpl_perm.
    rewrite In_mact,inverse_comp_l;apply H0.
  - apply eq_expr_eq;simpl;intuition.
    rewrite <-(rm_perm_apply_perm_swap _ _ );apply eq_expr_rm_perm.
    apply morphisms.apply_perm_equiv_perm;[|reflexivity].
    clear e H;intro x;repeat rewrite<-act_comp_app;simpl.
    case_eq_at x a;repeat simpl_beq;auto.
    case_eq_at a b;repeat simpl_beq;auto.
    now rewrite<- (perm_bij p) in *; repeat simpl_beq.
    case_eq_at x b;repeat simpl_beq;auto.
    now rewrite<- (perm_bij p) in *; repeat simpl_beq.
Qed.

Lemma apply_perm_θ_t e p :
  (exists A, A ⊨ e) ->
  eq_perm∨eq_perm_elim∨eq_perm_ω∨zero_perm_t ⊢ ‹ p › e ≡ p • e.
Proof.
  revert p;induction e;intros π (X&T);simpl in *;
  try destruct T as (T1 & T2); auto;try tauto.
  - now eql;eql;eqr;apply eq_ax, eq_θ_id.
  - now eql;eql;eqr;apply eq_ax, eq_θ_var.
  - now eql;eql;eqr;apply eq_ax, eq_θ_at.
  - rewrite <-IHe1,<-IHe2;(exists X;auto)||auto 10.
  - rewrite <-IHe1,<-IHe2;(exists X;auto)||auto 10.
  - rewrite <-IHe;(exists X;auto)||auto 10.
  - rewrite <-IHe;(exists (a::X);auto)||auto 10.
  - rewrite <-IHe;(exists (rm a X);auto)||auto 10.
  - rewrite <-IHe;(exists ((p⋆)⧓ X);auto)||auto 10.
    transitivity (‹π ++ p› e);[|symmetry];auto 10.
Qed.
Lemma apply_perm_θ_u e p :
  untyped e ->
  eq_perm∨eq_perm_elim∨zero_perm_u ⊢ ‹ p › e ≡ p • e.
Proof.
  revert p;induction e;intros π U;simpl in *;
  try destruct U as (U1 & U2); auto;try tauto.
  - rewrite <-IHe1,<-IHe2;auto 10.
  - rewrite <-IHe1,<-IHe2;auto 10.
  - rewrite <-IHe;auto 10.
  - rewrite <-IHe;auto 10.
  - rewrite <-IHe;auto 10.
    transitivity (‹π ++ p› e);[|symmetry];auto 10.
Qed.
    
Lemma eq_nom_untyped_np_apply_perm :
  forall e f , (ut ∧ eq_nom_untyped_np) e f ->
                 (eq_perm∨eq_perm_elim∨zero_perm_u∨
                  eq_nom_untyped_p) ⊢ e ≡ f.
Proof.
  intros e f (U&E); destruct E as [a b e F];
  destruct U as (U& U');simpl in *.
  etransitivity.
  unfold rel;eqr;apply eq_ax;apply (equ_θν' _ F);simpl;
  tauto.
  eapply eq_ν,stronger_ax_proof,apply_perm_θ_u;auto.
Qed.

Lemma eq_nom_typed_np_apply_perm :
  forall e f , (eq_nom_typed_np) e f ->
               (eq_perm∨eq_perm_elim∨ eq_perm_ω∨zero_perm_t∨
                         (eq_nom_typed_p)) ⊢ e ≡ f.
Proof.
  intros e f E;  destruct E as [A a b e T I].
  unfold rel in *; etransitivity.
  eqr;apply eq_ax;apply (eqt_θν' T I);simpl;tauto.
  eapply eq_ν,stronger_ax_proof,apply_perm_θ_t;auto;exists (a::A);
  simpl in *;tauto.
Qed.


Lemma eq_apply_perm a b :
  (forall p, Proper (a ==> eq b) (rm_perm ∘ (apply_perm p))) ->
  Proper (equiv_perm ==> eq a ==> eq b)
         (fun p e => rm_perm (p • e)).
Proof.
  unfold Basics.compose;
  intros hyp p q P e f E ;revert p q P;induction E;intros π π' P;
  simpl;auto;
  try (now apply (stronger_ax_proof (FFalse_spec _)),eq_rm_perm).
  - rewrite (IHE1 _ _ P);apply IHE2;reflexivity.
  - apply (stronger_ax_proof (FFalse_spec _));simpl;
    now rewrite H, (equiv_perm_mact _ _ P).
  - apply (stronger_ax_proof (FFalse_spec _));simpl;
    now rewrite H, (equiv_perm_mact _ _ P).
  - now apply eq_var; apply apply_perm_var_equiv_perm_eq_letter.
  - apply (stronger_ax_proof (FFalse_spec _));simpl.
    now rewrite (equiv_perm_spec _ P).
  - rewrite P;apply eq_ν,IHE;auto.
  - apply IHE;intros x;repeat rewrite <-act_comp_app;rewrite P,H;auto.
  - rewrite P;apply eq_ω,IHE;auto.
  - apply (stronger_ax_proof (FFalse_spec _));simpl.
    symmetry in P; apply eq_rm_perm;
    rewrite (apply_perm_equiv_perm _ P);clear a b hyp π' P.
    destruct H;simpl;auto.
  - simpl in IHE.
    destruct H;simpl in *;
    (etransitivity;
      [eapply eq_ka_impl;[|apply ka_s3||apply ka_s4];
      rewrite (IHE π π );[auto|reflexivity]|]);
    apply (stronger_ax_proof (FFalse_spec _));simpl;
    symmetry in P; apply eq_rm_perm;rewrite (apply_perm_equiv_perm _ P);
    reflexivity.
  - rewrite (hyp π _ _ H).
    apply (stronger_ax_proof (FFalse_spec _));simpl;
    symmetry in P; apply eq_rm_perm;rewrite (apply_perm_equiv_perm _ P);
    reflexivity.
Qed.

Lemma eq_rm_perm_ut_pos :
  forall (e : expr) (p : perm),
    untyped e ->  positive (‹ p › e) ->
    eq_perm ∨ eq_perm_elim⊢ ‹ p › e ≡ rm_perm (p • e).
Proof.
  induction e;intros q U1 P;simpl in *;try tauto;auto 10.
  rewrite <-IHe1,<-IHe2;auto;try tauto.
  rewrite <-IHe1,<-IHe2;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
Qed.

Lemma eq_rm_perm_pos :
  forall (e : expr) (p : perm),
     positive (‹ p › e) ->
    eq_perm_ω∨eq_perm ∨ eq_perm_elim⊢ ‹ p › e ≡ rm_perm (p • e).
Proof.
  induction e;intros q P;simpl in *;try tauto;auto 10.
  rewrite <-IHe1,<-IHe2;auto;try tauto.
  rewrite <-IHe1,<-IHe2;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
Qed.

Lemma eq_apply_perm_ut_pos :
  forall (e : expr) (p : perm),
    untyped e -> positive (‹ p › e) ->
    eq_perm ∨ eq_perm_elim⊢ ‹ p › e ≡ (p • e).
Proof.
  induction e;intros q U1 P;simpl in *;try tauto;auto 10.
  rewrite <-IHe1,<-IHe2;auto;try tauto.
  rewrite <-IHe1,<-IHe2;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  transitivity (‹[] ++q ++ p› e);auto;rewrite app_nil_l;auto.
Qed.

Lemma eq_apply_perm_pos :
  forall (e : expr) (p : perm),
    positive (‹ p › e) ->
    eq_perm_ω∨eq_perm ∨ eq_perm_elim⊢ ‹ p › e ≡ (p • e).
Proof.
  induction e;intros q P;simpl in *;try tauto;auto 10.
  rewrite <-IHe1,<-IHe2;auto;try tauto.
  rewrite <-IHe1,<-IHe2;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  rewrite <-IHe;auto;try tauto.
  transitivity (‹[] ++q ++ p› e);auto;rewrite app_nil_l;auto.
Qed.


Lemma eq_nom_untyped_np_rm_perm_pos :
  forall e f : expr, str_positive e -> str_positive f ->
                    (ut ∧ eq_nom_untyped_np) e f ->
                    eq_perm ∨ eq_perm_elim
                            ∨ eq_nom_untyped_p ⊢ e ≡ f.
Proof.
  intros e f P1 P2 ((U1&U2)&E).
  destruct E as [a b e F];simpl in *.
  etransitivity.
  eqr;apply eq_ax;apply (equ_θν' _ F).
  apply eq_ν;eql.
  rewrite eq_apply_perm_ut_pos;auto.
Qed.

