Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Import proofs.

Create HintDb propererase.
Lemma ka_eq_erase: Proper (ka_eq ==> ka_eq) erase.
Proof. intros ? ? E;destruct E;simpl;auto. Qed.
Lemma ka_eq_tu_erase: Proper (ka_eq_t ==> ka_eq_u) erase.
Proof. intros ? ? E;destruct E;simpl;auto. Qed.
Hint Resolve ka_eq_tu_erase ka_eq_erase : propererase.
Lemma eq_nom_erase:
  Proper (ty'∧ eq_nom_typed ==>
             eq eq_nom_untyped) erase.
Proof.
  intros e f ((A&_&Tf)&E);destruct E;simpl;auto 10.
  - simpl in *;simpl_beq;simpl in *;destruct H as (_&_&H).
    assert (F:~ In a (rm a A0)) by (simpl_in;tauto).
    apply (fresh_type H) in F;auto.
  - simpl in *;try simpl_beq;simpl in *.
    destruct Tf as (I&T1&_&T2).
    assert (F:~ In a (rm a A)) by (simpl_in;tauto).
    apply (fresh_type T2) in F;auto.
  - simpl in *;simpl_beq;simpl in *.
    destruct Tf as (I&(_&T1)&T2).
    assert (F:~ In a (rm a A)) by (simpl_in;tauto).
    apply (fresh_type T1) in F;auto.
Qed.
Lemma eq_nom_p_erase:
  Proper (ty'∧ eq_nom_typed_p ==>
             eq (eq_perm∨eq_perm_elim∨eq_nom_untyped_p)) erase.
Proof.
  intros e1 e2 ((A&_&Tf)&[X a b e [I T] J]).
  case_eq_at a b;simpl.
  eql;eqr;apply eq_ν;etransitivity;
  [|eapply eq_θ;[|reflexivity];symmetry;apply perm_a_a_eq_nil];
  auto.
  apply fresh_type with (a:= b) in T;[apply eq_ax|simpl;tauto].
  now right;apply equ_θν'.
Qed.
Hint Resolve eq_nom_erase eq_nom_p_erase : propererase.
Lemma eq_nom_untyped_np_no_perm:
  Proper (eq_nom_untyped_np ==> iff) no_perm.
Proof. intros e f E.
       pose proof (eq_nom_untyped_np_no_perm E).
       firstorder.
Qed.
       
Lemma eq_nom_np_erase:
  Proper ( np ∧ eq_nom_typed_np ==>
              eq (np ∧ eq_nom_untyped_np)) erase.
Proof.
  intros e1 e2 [[N1 N2] E].
  destruct E as [X a b e [I T] J];case_eq_at a b.
  - simpl in *; apply eq_ν,eq_expr_eq;symmetry.
    now apply eq_expr_erase; apply apply_perm_id. 
  - assert (F:b # ⎣e⎦) by (apply (fresh_type T);simpl;tauto).
    simpl;rewrite <-erase_apply_perm;prove_ax.
    unfold and_rel,np,sg,rel;split_conj;simpl in *;
    try apply no_perm_apply_perm; auto using no_perm_erase.
    now apply equ_θν.
Qed.
Hint Resolve eq_nom_np_erase : propererase.
Lemma eq_perm_elim_erase: Proper (eq_perm_elim ==> eq_perm_elim) erase.
Proof. intros ? ? E;destruct E;simpl;auto. Qed.
Hint Resolve eq_perm_elim_erase : propererase.
Lemma eq_perm_erase: Proper (eq_perm ==> eq_perm) erase.
Proof. intros ? ? E;destruct E;simpl;auto. Qed.
Hint Resolve eq_perm_erase : propererase.
Lemma eq_perm_ω_erase: Proper (eq_perm_ω ==> Logic.eq) erase.
Proof. intros ? ? E;destruct E;simpl;auto. Qed.
Lemma eq_ν_smpl_erase: Proper (eq_ν_smpl ==> eq_ν_smpl) erase.
Proof. intros ? ? E;destruct E;simpl;auto. Qed.
Lemma eq_ω_smpl_erase :
   Proper (eq_ω_smpl ==> Logic.eq) erase.
Proof. intros ? ? E;destruct E;simpl;auto. Qed.
Hint Resolve eq_perm_ω_erase eq_ω_smpl_erase eq_ν_smpl_erase
  : propererase.
Lemma eq_erase (ax1 ax2 : relation expr) :
  Proper (ty'∧ ax1 ==> eq ax2) erase ->
  (forall A, Proper (ax1 ==> iff) ( has_type A)) ->
  Proper (ty'∧ (eq ax1) ==> eq ax2) erase.
Proof.
  intros Hyp axtyp e f ((A&Te&Tf)&E).
  pose proof (@eq_type ax1 axtyp) as eqtyp.
  revert A Te Tf;induction E;simpl;
  intros X Te Tf;intuition eauto.
  - etransitivity;[eapply IHE1|eapply IHE2];try eassumption;
    rewrite (eqtyp X _ _ E2);auto.
  - rewrite <-(equiv_perm_mact _ _ (equiv_perm_inverse H)) in Tf;intuition eauto.
  - apply ka_eq_erase in H;auto.
  - destruct H;(eapply eq_ka_impl;
                 [rewrite <- IHE at 2;[simpl;reflexivity| |]|]);
    simpl in *;intuition eauto.
  - apply (Hyp e f);split;eauto.
Qed.

Create HintDb propersupport.
Lemma ka_eq_support :
  Proper (ka_eq ==> equiv_lst) support.
Proof.
  unfold equiv_lst;intros ? ? E ?;destruct E;
  simpl;simpl_in;intuition.
Qed.
Hint Resolve ka_eq_support : propersupport.
Lemma ka_eq_u_support :
  Proper (ka_eq_u ==> equiv_lst) support.
Proof.
  unfold equiv_lst;intros ? ? E ?;destruct E;
  simpl;simpl_in;intuition.
Qed.
Hint Resolve ka_eq_u_support : propersupport.

Lemma eq_nom_untyped_support :
  Proper (eq_nom_untyped ==> equiv_lst) support.
Proof.
  intros ? ? E ?;destruct E;simpl;simpl_in;simpl;
  apply fresh_support in H; intuition subst;tauto.
Qed.
Hint Resolve eq_nom_untyped_support : propersupport.
Lemma eq_nom_untyped_p_support :
  Proper (eq_nom_untyped_p ==> equiv_lst) support.
Proof.
  intros ? ? E ?;destruct E;simpl;simpl_in;simpl.
  apply fresh_support in H.
  case_eq_at x0 a;simpl_beq;[|case_eq_at x0 b; simpl_beq];
  intuition.
Qed.
Hint Resolve eq_nom_untyped_p_support : propersupport.
Lemma eq_nom_untyped_np_support :
  Proper (eq_nom_untyped_np ==> equiv_lst) support.
Proof.
  intros ? ? E ?;destruct E;simpl;simpl_in;simpl.
  apply fresh_support in H.
  rewrite (support_apply_perm e);simpl_in;simpl.
  case_eq_at x0 a;simpl_beq;[|case_eq_at x0 b; simpl_beq];
  intuition.
Qed.
Hint Resolve eq_nom_untyped_np_support : propersupport.


Lemma eq_perm_elim_support :
  Proper (eq_perm_elim ==> equiv_lst) support.
Proof.
  intros ? ? E;destruct E;simpl;auto;
  try rewrite H;simpl_perm;auto.
  now rewrite apply_perm_var_τ.
Qed.
Lemma eq_perm_support :
  Proper (eq_perm ==> equiv_lst) support.
Proof.
  unfold equiv_lst;intros ? ? E ?;destruct E;
  simpl;try (simpl_in;reflexivity).
  - simpl_in;intuition.
    now apply H1;rewrite <-H;simpl_perm.
    now apply H1;rewrite H;simpl_perm.
  - now rewrite mact_comp_app.
Qed.
Hint Resolve eq_perm_support eq_perm_elim_support : propersupport.
Lemma eq_ν_smpl_support :
  Proper (eq_ν_smpl ==> equiv_lst) support.
Proof.
  unfold equiv_lst;intros ? ? E ?;destruct E;
  simpl;simpl_in;intuition.
Qed.
Hint Resolve eq_ν_smpl_support : propersupport.

Lemma eq_support ax :
  Proper (ax ==> equiv_lst) support ->
  Proper (eq ax ==> equiv_lst) support.
Proof.
  intros Ax e f E;induction E;simpl;auto.
  - eauto.
  - rewrite eq_letter_τ;eauto.
  - rewrite IHE1,IHE2;reflexivity.
  - rewrite IHE1,IHE2;reflexivity.
  - rewrite IHE;reflexivity.
  - rewrite (equiv_perm_mact _ _ H),IHE;reflexivity.
  - now apply ka_eq_support.
  - rewrite<- IHE;destruct H;simpl;intro;simpl_in;tauto.
Qed.
Inductive eq_perm_W : relation expr :=
| eq_θ_W p a e : eq_perm_W (‹ p › (W a e)) (W (p⧓a) (‹ p › e)).
Inductive eq_W_smpl : expr -> expr -> Prop :=          
| eq_W_id a A : NoDup a -> a∩A = [] ->
                eq_W_smpl (W a (id A)) (id (a++A))
| eq_W_p e f a : eq_W_smpl (W a (e + f)) (W a e + W a f)
| eq_W_t e f a : eq_W_smpl (W a (e · f)) (W a e · W a f)
| eq_W_s e a : eq_W_smpl (W a (e^* )) ((W a e)^* )
| eq_WW e a b : eq_W_smpl (W a (W b e)) (W b (W a e)).
Lemma eq_perm_ω_W :
  eq_perm_W ⇒  eq eq_perm_ω.
Proof.
  intros e f E;destruct E.
  induction a;simpl;auto.
  rewrite <-IHa;eauto.
Qed.
Lemma Ty_eq_perm_ω_W :
   (ty' ∧ eq_perm_W) ⇒ eq eq_perm_ω.
Proof.
  intros e f [[A [T1 T2]] E];destruct E.
  revert A T1 T2;induction a;intros A T1 T2;simpl;auto.
  rewrite <-(IHa (rm (p⋈a) A));[apply eq_ax;split;auto;exists A| |];
  simpl in *;rewrite mact_rm in *;simpl_perm;
  repeat split; tauto.
Qed.
Global Instance W_eq ax a : Proper (eq ax ==> eq ax) (W a).
Proof. induction a;intros ? ? E;simpl;auto. Qed.

Lemma W_ω_exchange :
  forall a b e, eq_ω_smpl ⊢ ω a (W b e) ≡ W b (ω a e).
Proof. intros a b e;induction b;simpl;auto. rewrite <-IHb;auto. Qed.

Lemma eq_smpl_ω_W :
  forall e f, eq_W_smpl e f ->  eq_ω_smpl ⊢ e ≡ f.
Proof.
  intros e f E;destruct E;
  try (now induction a;simpl;auto;simpl;rewrite IHa;eauto).
  - induction a;simpl;auto.
    assert ((a0 ∩ A) = [] /\ ~ In a A) as (C&IA).
    simpl in H0; case_ex a A;[discriminate|tauto].
    apply NoDup_cons_iff in H as (Ia0&ND).
    rewrite IHa,app_comm_cons;auto.
    apply eq_ax,eq_ω_id;simpl_in;tauto.
  - revert b e;induction a;intros b e;simpl;auto.
    rewrite <-W_ω_exchange;apply eq_ω;auto.
Qed.

Lemma W_equiv_lst l m e : NoDup l -> NoDup m -> l ≈ m ->
                          eq_ω_smpl ⊢ W l e ≡ W m e.
Proof.
  revert m;induction l;intros m NDl NDm E.
  - destruct m;simpl;auto.
    assert (In a [])by (now rewrite E;left).
    simpl in H;tauto.
  - assert (In a m)by (now rewrite <-E;left).
    apply in_split in H as (m1 & m2 & ->).
    rewrite NoDup_cons_iff in NDl;
      pose proof (NoDup_remove_1 _ _ _ NDm) as NDm';
      pose proof (NoDup_remove_2 _ _ _ NDm) as Im.
    rewrite <-W_app;simpl.
    rewrite <-W_ω_exchange,W_app;apply eq_ω,IHl;try tauto.
    destruct NDl as (Il&NDl);simpl_in.
    rewrite <- (cons_rm_eq' Il),E,rm_app;simpl;simpl_beq;
    repeat rewrite rm_id;reflexivity||tauto.
Qed.
Lemma eq_νW2 :
  forall e a l, ~ In a l -> NoDup l ->
                 eq_nom_typed ⊢ ν a (W l e) ≡ W l (ν a e).
Proof.
  intros e a l Ia ND;induction l;simpl;auto.
  simpl in Ia;apply NoDup_cons_iff in ND as (Ia0&ND).
  rewrite<-IHl;auto;clear IHl.
  case_eq_at a0 a;[tauto|].
  prove_ax.
Qed.
Ltac prove_eq_W :=
  repeat
    match goal with
    | _:_ |- _ ⊢ _·_ ≡ _·_ => apply eq_times
    | |- _ ⊢ _·_ ≡ _·_ => apply eq_times
    | _:_ |- _ ⊢ _+_ ≡ _+_ => apply eq_plus
    | |- _ ⊢ _+_ ≡ _+_ => apply eq_plus
    | _:_ |- _ ⊢ _^* ≡ _^* => apply eq_star
    | |- _ ⊢ _^* ≡ _^* => apply eq_star
    | _:_ |- _ ⊢ ν _ _ ≡ ν _ _ => apply eq_ν
    | |- _ ⊢ ν _ _ ≡ ν _ _ => apply eq_ν
    | _:_ |- _ ⊢ ω _ _ ≡ ω _ _ => apply eq_ω
    | |- _ ⊢ ω _ _ ≡ ω _ _ => apply eq_ω
    | _:_ |- _ ⊢ ‹ _ ›_ ≡ ‹ _ › _ => apply eq_θ
    | |- _ ⊢ ‹ _ ›_ ≡ ‹ _ › _ => apply eq_θ
    | _:_ |- _ ⊢ W _ (_·_) ≡ _ =>
      etransitivity;[apply eq_smpl_ω_W;apply eq_W_t|]
    | |- _ ⊢ W _ (_·_) ≡ _ =>
      etransitivity;[apply eq_smpl_ω_W;apply eq_W_t|]
    | _:_ |- _ ⊢ W _ (_+_) ≡ _ =>
      etransitivity;[apply eq_smpl_ω_W;apply eq_W_p|]
    | |- _ ⊢ W _ (_+_) ≡ _ =>
      etransitivity;[apply eq_smpl_ω_W;apply eq_W_p|]
    | _:_ |- _ ⊢ W _ (_^* ) ≡ _ =>
      etransitivity;[apply eq_smpl_ω_W;apply eq_W_s|]
    | |- _ ⊢ W _ (_^* ) ≡ _ =>
      etransitivity;[apply eq_smpl_ω_W;apply eq_W_s|]
    | _:_ |- _ ⊢ W _ (‹ _ ›_) ≡ _ =>
      etransitivity;[apply eq_perm_ω_W;apply eq_θ_W|]
    | |- _ ⊢ W _ (‹ _ ›_) ≡ _ =>
      etransitivity;[apply eq_perm_ω_W;apply eq_θ_W|]
    | _:_ |- _ ⊢ W _ (W _ _) ≡ _ =>
      rewrite W_app
    | |- _ ⊢ W _ (W _ _) ≡ _ =>
      rewrite W_app
    | _:_ |- _ ⊢ W _ _ ≡ W _ _ =>
      eapply W_equiv_lst;repeat solve_no_dup;try intro;
      simpl_in;intuition
    | |- _ ⊢ W _ _ ≡ W _ _ =>
      eapply W_equiv_lst;repeat solve_no_dup;try intro;
      simpl_in;intuition
    end.

Lemma W_ω_extract :
  forall a l e, NoDup l -> In a l -> 
                exists m, NoDup m /\ rm a l ≈ m
                          /\ eq_ω_smpl ⊢ W l e ≡ ω a (W m e).
Proof.
  intros a l e ND I.
  apply in_split in I as (m&k&h);rewrite h in *.
  pose proof (NoDup_remove_1 _ _ _ ND) as ND';
  pose proof (NoDup_remove_2 _ _ _ ND) as I.
  exists (m++k);split_conj;auto.
  rewrite rm_app;simpl;simpl_beq;simpl;rewrite<-rm_app,rm_id;auto.
  repeat rewrite <-W_app;simpl.
  symmetry;apply W_ω_exchange.
Qed.
Create HintDb propertypedvers.
Lemma ka_eq_typed_vers:
  Proper (ut ∧ ka_eq ==> eq eq_ω_smpl) typed_vers.
Proof.
  intros ? ? (T&E);destruct E;simpl;auto.
  - destruct T as ((Ue & Uf & Ug)&_);
    pose proof (typed_version_support_type Ue) as Te;
    pose proof (typed_version_support_type Uf) as Tf;
    pose proof (typed_version_support_type Ug) as Tg.
    etransitivity;
    [|etransitivity;[apply eq_ka,ka_tt|symmetry]].
    prove_eq_W; try reflexivity.
    prove_eq_W;try eassumption
               || (apply disjoint_In;intro;simpl_in;intuition)
               || (case_in x (support e);tauto)
               || (case_in x (support f);tauto)
               || (case_in x (support g);tauto). 
  - destruct T as ((Ue & Uf & Ug)&_);
    pose proof (typed_version_support_type Ue) as Te;
    pose proof (typed_version_support_type Uf) as Tf;
    pose proof (typed_version_support_type Ug) as Tg.
    etransitivity;
    [|etransitivity;[apply eq_ka,ka_pp|symmetry]].
    prove_eq_W; try reflexivity.
    prove_eq_W;try eassumption
               || (apply disjoint_In;intro;simpl_in;intuition)
               || (case_in x (support e);tauto)
               || (case_in x (support f);tauto)
               || (case_in x (support g);tauto).
  - repeat rewrite <-undup_purge,purge_l_l;simpl;auto.
  - destruct T as ((Ue & Uf & Ug)&_);
    pose proof (typed_version_support_type Ue) as Te;
    pose proof (typed_version_support_type Uf) as Tf;
    pose proof (typed_version_support_type Ug) as Tg.
    etransitivity;
    [|etransitivity;[apply eq_ka,ka_pt|symmetry]].
    prove_eq_W; try reflexivity.
    prove_eq_W;try eassumption
               || (apply disjoint_In;intro;simpl_in;intuition)
               || (case_in x (support e);tauto)
               || (case_in x (support f);tauto)
               || (case_in x (support g);tauto).
  - destruct T as (((Ue & Uf) & Ug)&_);
    pose proof (typed_version_support_type Ue) as Te;
    pose proof (typed_version_support_type Uf) as Tf;
    pose proof (typed_version_support_type Ug) as Tg.
    etransitivity;
    [|etransitivity;[apply eq_ka,ka_tp|symmetry]].
    prove_eq_W; try reflexivity.
    prove_eq_W;try eassumption
               || (apply disjoint_In;intro;simpl_in;intuition)
               || (case_in x (support e);tauto)
               || (case_in x (support f);tauto)
               || (case_in x (support g);tauto).
Qed.
Hint Resolve typed_version_support_type.
Lemma ka_eq_tu_typed_vers:
  Proper (ut ∧ ka_eq_u ==> eq (ka_eq_t∨eq_ω_smpl)) typed_vers.
Proof.
  pose proof (stronger_ax_proof (@or_rel_stronger_l _ ka_eq_t eq_ω_smpl))
    as S1.
  pose proof (stronger_ax_proof (@or_rel_stronger_r _ ka_eq_t eq_ω_smpl))
    as S2.
  intros ? ? (U&E);unfold ut,and_rel,rel in *;
  destruct E;simpl;simpl_hyp;auto;
  rewrite purge_nil.
  - etransitivity;[apply S2|apply S1,eq_ax,kat_t1;eauto].
    etransitivity;[apply eq_times;[auto|];apply eq_smpl_ω_W;
                   apply eq_W_id;[solve_no_dup
                                 |apply disjoint_In;simpl;tauto]|].
    rewrite app_nil_r;auto.
    apply eq_times;reflexivity||apply eq_id,undup_equiv_lst.
  - etransitivity;[apply S2|apply S1,eq_ax, kat_1t;eauto].
    apply eq_times;auto.
    etransitivity;[apply eq_smpl_ω_W;apply eq_W_id;
                   [solve_no_dup|apply disjoint_In;simpl;tauto]|].
    rewrite app_nil_r;apply eq_id,undup_equiv_lst.
  - repeat rewrite<- undup_purge;repeat rewrite purge_l_m_nil;
    try rewrite app_l_l;try reflexivity;simpl in *.
    rewrite undup_app,purge_l_l;simpl.
    etransitivity.
    apply eq_plus;[apply eq_plus;[apply S2|reflexivity]|reflexivity].
    etransitivity;[apply eq_smpl_ω_W;apply eq_W_id;
                   [solve_no_dup|apply disjoint_In;simpl;tauto]|].
    rewrite app_nil_r;auto.
    assert (T : 〈support e〉⊨⎡e⎤)
      by (rewrite undup_equiv_lst;apply (typed_version_support_type h0)).
    apply S1,eq_ax, (kat_s1 T).
  - repeat rewrite<- undup_purge;repeat rewrite purge_l_m_nil;
    try rewrite app_l_l;try reflexivity;simpl in *.
    rewrite undup_app,purge_l_l;simpl.
    etransitivity.
    apply eq_plus;[apply eq_plus;[apply S2|reflexivity]|reflexivity].
    etransitivity;[apply eq_smpl_ω_W;apply eq_W_id;
                   [solve_no_dup|apply disjoint_In;simpl;tauto]|].
    rewrite app_nil_r;auto.
    assert (T : 〈support e〉⊨⎡e⎤)
      by (rewrite undup_equiv_lst;apply (typed_version_support_type h0)).
    apply S1,eq_ax,(kat_s2 T).
Qed.
Hint Resolve ka_eq_typed_vers ka_eq_tu_typed_vers : propertypedvers.
Lemma eq_nom_typed_vers:
  Proper (ut ∧ eq_nom_untyped ==>
             eq (eq_ω_smpl∨eq_nom_typed))
         typed_vers.
Proof.
  pose proof (stronger_ax_proof
                (@or_rel_stronger_l _ eq_ω_smpl eq_nom_typed)) as S1.
  pose proof (stronger_ax_proof
                (@or_rel_stronger_r _ eq_ω_smpl eq_nom_typed)) as S2.
  intros e f ((_&Tf)&E);destruct E;simpl in *;auto.
  - case_ex a (support e).
    apply fresh_support in H;tauto.
    eapply S2,eq_ax, eqt_νω1;simpl;simpl_beq;simpl;
    split_conj;try rewrite rm_id;eauto.
  - destruct Tf as (Uf & Ue);
    pose proof (typed_version_support_type Ue) as Te;
    pose proof (typed_version_support_type Uf) as Tf.
    assert (〈support f ++ support e〉 ⊨ W (〈support e〉 -- support f) ⎡f⎤
            /\
            〈support f ++ support e〉 ⊨ W (〈support f〉 -- support e) ⎡e⎤)
      as (h1 & h2).
    split;(apply W_type;[prove_type| |repeat solve_no_dup]);
    intro x;case_eq_at x a;simpl_in;intuition.
    apply fresh_support in H.
    case_ex a (support e);[tauto|].
    case_ex a (support f);
      case_ex a (support f ++ support e);try (simpl_in;tauto).
    * clear H;rewrite <-undup_purge,purge_rm;auto;etransitivity.
       apply eq_times;[|reflexivity].
       symmetry;eapply S2,eq_νW2;repeat solve_no_dup;simpl_in;tauto.
       etransitivity;[apply S2,eq_ax,eqt_ν3|].
       apply eq_ν,eq_times,S1;auto.
       now rewrite undup_purge.
       destruct (@W_ω_extract a (〈support f〉 -- support e) (⎡e⎤))
         as (m & h4 & h5 & Eq);[| |rewrite Eq in *];
       repeat solve_no_dup;simpl_in;auto.
       rewrite<- undup_purge,rm_purge; apply eq_ω,W_equiv_lst;
       auto;repeat solve_no_dup.
       rewrite undup_equiv_lst in *;auto.
    * rewrite rm_id;auto.
       etransitivity;[|symmetry;apply S2,eq_ax,
                                (@eqt_νω1 〈support f ++ support e〉);
                       simpl;simpl_beq;simpl;
                       rewrite rm_id;simpl_in;firstorder].
       assert (Eq: eq_ω_smpl ∨ eq_nom_typed
                                 ⊢ν a (ω a (⎡f⎤)) ≡ ⎡f⎤);
         [|rewrite Eq;auto].
       eapply S2,eq_ax, (@eqt_νω1 (support f)). 
       simpl;simpl_beq;simpl;erewrite (rm_id h0);auto.
  - destruct Tf as (Uf & Ue);
    pose proof (typed_version_support_type Ue) as Te;
    pose proof (typed_version_support_type Uf) as Tf.
    assert ((support f ++ support e) ⊨ W (〈support e〉 -- support f) ⎡f⎤
            /\
            (support f ++ support e) ⊨ W (〈support f〉 -- support e) ⎡e⎤)
      as (h1 & h2).
    split;(apply W_type;[prove_type| |repeat solve_no_dup]);
    intro x;case_eq_at x a;simpl_in;intuition.
    case_ex a (support e).
    apply fresh_support in H;tauto.
    case_ex a (support f);
      case_ex a (support e++support f);try (simpl_in;tauto).
    * repeat rewrite <-undup_purge; rewrite purge_rm;auto;
       etransitivity.
       apply eq_times;[reflexivity|].
       symmetry;apply S2,eq_νW2;repeat solve_no_dup;simpl_in;tauto.
       etransitivity;[apply S2,eq_ax,eqt_ν4
                     |apply S1,eq_ν,eq_times;auto].
       destruct (@W_ω_extract a 〈support f -- support e〉 (⎡e⎤))
         as (m & h4 & h5 & Eq);[| |rewrite Eq in *];
       auto;repeat solve_no_dup;
       simpl_in;auto.
       rewrite rm_purge; apply eq_ω,W_equiv_lst;auto;
       repeat solve_no_dup.
       now repeat rewrite undup_equiv_lst in *.
    * rewrite rm_id;auto.
       etransitivity;[|symmetry;
                        apply S2,eq_ax,(@eqt_νω1 (support f ++ support e));
                        simpl;simpl_beq;repeat rewrite rm_id;
                        simpl_in;firstorder].
       assert (Eq:eq_ω_smpl ∨ eq_nom_typed
                                ⊢  ν a (ω a (⎡f⎤)) ≡ ⎡f⎤);
         [|rewrite Eq in *;auto].
       eapply S2,eq_ax, (@eqt_νω1 (support f)).
       simpl;simpl_beq;simpl;rewrite (rm_id h0);auto.
Qed.
Hint Resolve eq_nom_typed_vers : propertypedvers.
Lemma eq_nom_typed_p_vers:
  Proper (ut ∧ eq_nom_untyped_p ==>
             eq (eq_perm_ω∨eq_perm
                          ∨eq_perm_elim∨eq_nom_typed_p))
         typed_vers.
Proof.
  intros e f ((_&U)&E);destruct E;simpl in *;auto.
  pose proof (typed_version_support_type U) as T.
  apply fresh_support in H.
  case_ex b (support e);
    case_ex a ([a, b]  ⧓ (support e));
    simpl_in;simpl in *;simpl_beq;[|tauto|tauto|].
  - case_eq_at b a.
    * eql;eqr;apply eq_ν;symmetry;etransitivity;
       [apply eq_θ;[|reflexivity];apply perm_a_a_eq_nil|auto].
    * apply eq_ax;right; apply (@eqt_θν' (rm b (support e)));
       simpl;try rewrite <-(rm_cons_eq' h0);simpl_in;tauto.
  - etransitivity.
    apply eq_ax;right; eapply (@eqt_θν' _ _ a);simpl;eauto;
    simpl;repeat (simpl_beq;simpl);try rewrite (mact_id2 H n);
    rewrite (rm_id h);tauto.
    apply eq_ν; etransitivity.
    apply eq_ax;left;left;left;apply eq_θ_ω.
    case_eq_at b a; simpl;repeat simpl_beq;auto 10.
Qed.
Hint Resolve eq_nom_typed_p_vers : propertypedvers.
Lemma eq_nom_typed_np_vers:
  Proper (np ∧ (ut ∧ eq_nom_untyped_np) ==>
             eq (np ∧ (eq_nom_typed_np∨eq_nom_typed)))
         typed_vers.
Proof.
  unfold np,and_rel, ut,ty', rel.
  intros e f ((NP & _) & (U&_)&E);
  destruct E as [a b e I];simpl in *;auto.
  apply fresh_support in I.
  pose proof (typed_version_support_type U) as T.
  repeat rewrite <-(typed_vers_apply_perm ).
  rewrite (support_apply_perm);auto.
  case_ex b (support e);
    case_ex a ([a, b]  ⧓ (support e));
    simpl_in;simpl in *;simpl_beq;[|tauto|tauto|].
  - case_eq_at a b.
    * symmetry; apply eq_ν,eq_expr_eq, apply_perm_id;
      auto using no_perm_typed_vers.
    * apply eq_ax;split_conj;simpl in *;
       try apply no_perm_apply_perm;
       auto using no_perm_typed_vers.
    left;apply (@eqt_θν (rm b (support e)));simpl;simpl_in;
      try rewrite <-rm_cons_eq';intuition.
  - transitivity (ν a ([a, b] • ω b (⎡e⎤))).
    * apply eq_ax;split_conj;
       try apply no_perm_apply_perm;
       try (now simpl;auto using no_perm_typed_vers).
       left; eapply (@eqt_θν _ _ a);simpl;eauto.
       simpl_beq;simpl;rewrite rm_id;split_conj;eauto.
    * case_eq_at b a; simpl;repeat simpl_beq;auto 10.
Qed.
Hint Resolve eq_nom_typed_np_vers : propertypedvers.

Lemma eq_perm_elim_typed_vers:
  Proper (eq_perm_elim ==> eq_perm_elim) typed_vers.
Proof. intros ? ? E;destruct E;simpl;auto. Qed.
Hint Resolve eq_perm_elim_typed_vers : propertypedvers.
Lemma eq_perm_typed_vers:
  Proper (ut ∧ eq_perm ==> eq (eq_nom_typed∨eq_perm_ω∨eq_perm))
         typed_vers.
Proof.
  intros ? ? (U&E);destruct E;simpl;auto.
  - repeat rewrite <-undup_purge;
    repeat rewrite mact_purge,<-mact_undup;simpl_perm.
    etransitivity.
    apply eq_ax;right;apply eq_θ_p.
    apply (@stronger_ax_proof eq_perm_ω).
    intro;unfold or_rel;intuition.
    apply eq_plus;apply eq_perm_ω_W,eq_θ_W .
  - repeat rewrite <-undup_purge;
    repeat rewrite mact_purge,<-mact_undup;simpl_perm.
    etransitivity.
    apply eq_ax;right;apply eq_θ_t.
    apply (@stronger_ax_proof eq_perm_ω).
    intro;unfold or_rel;intuition.
    apply eq_times;apply eq_perm_ω_W, eq_θ_W .
  - case_ex a (support e);case_ex (p⋈a) (p⧓support e);
    try (simpl_in;simpl_perm;tauto);auto.
    destruct U as (U1 & U2);simpl in *.
    etransitivity.
    apply (@eq_θ _ _ p);[intuition|prove_ax]. eapply eqt_νω1.
    simpl;simpl_beq;simpl;split_conj;try prove_type;
    try rewrite rm_id;eauto.
    symmetry;prove_ax. eapply eqt_νω1.
    simpl;simpl_beq;simpl;split_conj;try prove_type;
    try rewrite rm_id;eauto;simpl_perm;auto.
Qed.
Hint Resolve eq_perm_typed_vers : propertypedvers.
Lemma eq_ν_smpl_typed_vers:
  Proper (ut ∧ eq_ν_smpl ==> eq (eq_nom_typed∨eq_ω_smpl∨eq_ν_smpl))
         typed_vers.
Proof.
  assert (S1:forall e f, eq_nom_typed ⊢ e ≡ f ->
                         (eq_nom_typed∨eq_ω_smpl∨eq_ν_smpl) ⊢ e ≡ f)
    by (intros e f;apply stronger_ax_proof;unfold or_rel,stronger;
        intuition).
  assert (S2:forall e f, eq_ω_smpl ⊢ e ≡ f ->
                         (eq_nom_typed∨eq_ω_smpl∨eq_ν_smpl) ⊢ e ≡ f)
    by (intros e f;apply stronger_ax_proof;unfold or_rel,stronger;
        intuition).
  assert (S3:forall e f, eq_ν_smpl ⊢ e ≡ f ->
                         (eq_nom_typed∨eq_ω_smpl∨eq_ν_smpl) ⊢ e ≡ f)
    by (intros e f;apply stronger_ax_proof;unfold or_rel,stronger;
        intuition).
  intros ? ? (U&E);destruct E;simpl;auto.
  - simpl in U;destruct U as ((Ue & Uf)& _ & _).
    assert (T1 : (support f ++ support e)
                   ⊨ W (〈support f〉 -- support e) ⎡e⎤).
    apply W_type;try prove_type;repeat solve_no_dup;intro x;simpl_in;
    case_in x (support e);tauto.
    assert (T2 : (support f ++ support e)
                   ⊨ W (〈support e〉 -- support f) ⎡f⎤).
    apply W_type;try prove_type;repeat solve_no_dup;intro x;simpl_in;
    case_in x (support f);tauto.
    pose proof (typed_version_support_type Ue) as Te.
    pose proof (typed_version_support_type Uf) as Tf.
    case_ex a (support e ++ support f);
      case_ex a (support e);
      case_ex a (support f);
      try (simpl_in;tauto).
    * repeat rewrite <-undup_purge.
       repeat rewrite purge_rm,rm_purge,rm_id;simpl_in;try tauto.
       etransitivity;
         [|apply S1;repeat rewrite <- eq_νW2;try reflexivity];
            auto;repeat solve_no_dup;simpl_in;tauto.
    * repeat rewrite <-undup_purge.
       rewrite purge_rm,rm_purge,rm_id;simpl_in;try tauto.
       rewrite (rm_id h1),rm_purge.
       assert (Eq: ((eq_nom_typed ∨ eq_ω_smpl) ∨ eq_ν_smpl)
                     ⊢ ν a (ω a ⎡f⎤) ≡ ⎡f⎤) by
           (apply S1,eq_ax,(@eqt_νω1 (support f));simpl;
            simpl_beq;simpl;rewrite rm_id;simpl_in;tauto);
       rewrite Eq in *;auto.
       etransitivity;
         [|apply S1;repeat rewrite <- eq_νW2;try reflexivity];
         auto;repeat solve_no_dup;simpl_in;try tauto.
       destruct (@W_ω_extract a 〈support e -- support f〉 (⎡f⎤))
         as (m&ND&L&E);
         repeat solve_no_dup;simpl_in;auto.
       rewrite undup_rm.
       etransitivity;[eqr;apply eq_ax,eq_νp
                     |apply eq_plus;[reflexivity|]].
       etransitivity;[eql;eqr;rewrite E;reflexivity|].
       etransitivity;[eql;eql;apply eq_ax,eqt_νω1
                              with (A:=rm a (support f ++ support e))|].
       simpl;simpl_beq;simpl_in;rewrite rm_id;simpl_in;firstorder.
       apply W_type;auto; rewrite <-L;
       [eapply ht_convert;[|apply typed_version_support_type;auto]|];
       rewrite undup_equiv_lst;
       intro x;simpl_in;case_eq_at x a;intuition.
       eql;eqr;apply W_equiv_lst;repeat solve_no_dup;auto.
    * repeat rewrite (rm_id h0).
       rewrite undup_rm,rm_purge, purge_rm;simpl_in;auto. 
       assert (Eq1: ((eq_nom_typed ∨ eq_ω_smpl) ∨ eq_ν_smpl)
                      ⊢ ν a (ω a ⎡e⎤) ≡ ⎡e⎤) by
           (apply S1,eq_ax,(@eqt_νω1 (support e));simpl;
            simpl_beq;simpl;rewrite rm_id;simpl_in;tauto);
       rewrite Eq1 in *.
       etransitivity;
         [|apply S1;repeat rewrite <- eq_νW2;try reflexivity];
         auto;repeat solve_no_dup;simpl_in;try tauto.
       destruct (@W_ω_extract a (〈support f〉 -- support e) (⎡e⎤))
         as (m&ND&L&E);
         repeat solve_no_dup;simpl_in;auto.
       etransitivity;[eqr;apply eq_ax,eq_νp
                     |apply eq_plus;[|reflexivity]].
       etransitivity;[eql;eqr;rewrite E;reflexivity|].
       etransitivity;[eql;eql;apply eq_ax,eqt_νω1
                              with (A:=rm a (support f ++ support e))|].
       simpl;simpl_beq;simpl_in;rewrite rm_id;simpl_in;firstorder.
       apply W_type;auto; rewrite <-L;
       [eapply ht_convert;[|apply typed_version_support_type;auto]|];
       rewrite undup_equiv_lst;
       intro x;simpl_in;case_eq_at x a;intuition.
       eql;eqr;apply W_equiv_lst;repeat solve_no_dup;auto.
    * setoid_replace (ν a (ω a ⎡e⎤))
      with (⎡e⎤) using relation
                 (eq ((eq_nom_typed ∨ eq_ω_smpl) ∨ eq_ν_smpl))
        by (apply S1,eq_ax,(@eqt_νω1 (support e));simpl;
            simpl_beq;simpl;rewrite rm_id;simpl_in;tauto).
      setoid_replace (ν a (ω a ⎡f⎤))
      with (⎡f⎤) using relation
                 (eq ((eq_nom_typed ∨ eq_ω_smpl) ∨ eq_ν_smpl))
        by (apply S1,eq_ax,(@eqt_νω1 (support f));simpl;
            simpl_beq;simpl;rewrite rm_id;simpl_in;tauto).
       setoid_replace
         (ν a (ω a (W ( 〈support f〉 -- support e) ⎡e⎤ +
                    W ( 〈support e〉 -- support f) ⎡f⎤)))
       with (W (〈support f〉--support e) ⎡e⎤ +
             W (〈support e〉--support f) ⎡f⎤)
              using relation
              (eq ((eq_nom_typed ∨ eq_ω_smpl) ∨ eq_ν_smpl))
        by (apply S1,eq_ax,(@eqt_νω1 (support f ++ support e));simpl;
            simpl_beq;simpl;rewrite rm_id;simpl_in;tauto).
       repeat rewrite rm_id;auto.
  - simpl in U;destruct U as (U&_).
    case_ex a (support e);case_ex b (support e);
    case_ex a (rm b (support e));
    case_ex b (rm a (support e));
    simpl_in;try (tauto);auto.
    * exfalso;apply h2;split;[|intros ->];tauto.
    * exfalso;apply h1;split;[|intros ->];tauto.
    * case_eq_at a b;[|tauto];auto.
    * etransitivity;[apply eq_ν|symmetry];
       prove_ax;eapply eqt_νω1;simpl;simpl_beq;simpl;
       split_conj;try (now left);
       try (prove_type;rewrite rm_id;try reflexivity);auto;
       simpl_in;tauto.
    * exfalso;apply h1;split;[|intros ->];tauto.
    * etransitivity;[|symmetry;apply eq_ν];
       prove_ax;eapply eqt_νω1;simpl;simpl_beq;simpl;
       split_conj;try (now left);
       try (prove_type;rewrite rm_id;try reflexivity);auto;
       simpl_in;tauto.
    * exfalso;apply h2;split;[|intros ->];tauto.
    * etransitivity.
       prove_ax;eapply eqt_νω1;simpl;simpl_beq;simpl;
       split_conj;try (now left);
       try (prove_type;rewrite rm_id;try reflexivity);auto;
       simpl;try (simpl_beq;simpl;rewrite rm_id);try reflexivity;
       simpl_in;tauto.
       etransitivity.
       prove_ax;eapply eqt_νω1;simpl;simpl_beq;simpl;
       split_conj;try (now left);
       try (prove_type;rewrite rm_id;try reflexivity);auto;
       simpl;try (simpl_beq;simpl;rewrite rm_id);try reflexivity;
       simpl_in;tauto.
       symmetry;etransitivity.
       prove_ax;eapply eqt_νω1;simpl;simpl_beq;simpl;
       split_conj;try (now left);
       try (prove_type;rewrite rm_id;try reflexivity);auto;
       simpl;try (simpl_beq;simpl;rewrite rm_id);try reflexivity;
       simpl_in;tauto.
       etransitivity.
       prove_ax;eapply eqt_νω1;simpl;simpl_beq;simpl;
       split_conj;try (now left);
       try (prove_type;rewrite rm_id;try reflexivity);auto;
       simpl;try (simpl_beq;simpl;rewrite rm_id);try reflexivity;
       simpl_in;tauto.
       auto.
Qed.
Hint Resolve eq_ν_smpl_typed_vers : propertypedvers.
Lemma eq_typed_vers (ax1 ax2 : relation expr) :
  Proper (ut ∧ ax1 ==> eq ax2) typed_vers ->
  Proper (ax1 ==> iff) untyped ->
  Proper (ax1 ==> equiv_lst) support ->
  eq_ω_smpl ⇒ ax2 ->
  Proper (ut ∧ (eq ax1) ==> eq ax2) typed_vers.
Proof.
  intros Hyp axtyp hsupport eqω e f ((Te&Tf)&E).
  pose proof (@eq_untyped ax1 axtyp) as eqtyp.
  pose proof (@eq_support ax1 hsupport) as eqsupport.
  revert Te Tf;induction E;simpl;
  intros Te Tf;intuition.
  - rewrite <- (eqtyp _ _ E1) in *;eauto.
  - rewrite H5,H3; apply eqsupport in E1; apply eqsupport in E2.
    apply (stronger_ax_proof eqω);apply eq_plus;apply W_equiv_lst;
    repeat solve_no_dup;rewrite E1, E2;reflexivity.
  - rewrite H5,H3; apply eqsupport in E1; apply eqsupport in E2.
    apply (stronger_ax_proof eqω);apply eq_times;apply W_equiv_lst;
    repeat solve_no_dup;rewrite E1, E2;reflexivity.
  - apply eqsupport in E;case_ex a (support e);
    case_ex a (support f);
    try (rewrite E in h;tauto);auto.
  - apply (stronger_ax_proof eqω),ka_eq_typed_vers;split;auto.
  - apply eqsupport in E; repeat rewrite <- undup_purge in *.
    destruct H;simpl in *;intuition; repeat rewrite <- undup_purge in *;
    repeat rewrite purge_l_m_nil in H3
      by (try rewrite <- E;intro;simpl_in;intuition);simpl in H3.
    rewrite (@purge_l_m_nil (_ ++ _ ++ _)) in H3
      by (rewrite <- E at 2;intro;simpl_in;intuition);simpl in H3.
    rewrite (@purge_l_m_nil (support f)) in H3
      by (try rewrite <- E;intro;simpl_in;intuition);
    rewrite (@purge_l_m_nil (support e)) in H3
      by (try rewrite <- E;intro;simpl_in;intuition);
      simpl in H3.
    rewrite (@purge_l_m_nil (support e ++ support f)) 
      by (rewrite <- E;intro;simpl_in;intuition);simpl.
    etransitivity;[|eapply eq_ka_impl;try apply ka_s3;apply H3].
    apply eq_plus;auto.
    apply (stronger_ax_proof eqω).
    etransitivity.
    apply eq_smpl_ω_W,eq_W_t.
    apply eq_times;auto;rewrite W_app.
    etransitivity.
    apply eq_smpl_ω_W,eq_W_s.
    apply eq_star,W_equiv_lst;repeat solve_no_dup.
    intro;simpl_in;intuition;try rewrite<-E;simpl_in;tauto.
    rewrite<-E at 2;intro;simpl_in;intuition;
    case_ex x (support f);tauto.
    apply W_equiv_lst;repeat solve_no_dup.
    intro;simpl_in;intuition;try rewrite<-E;simpl_in;tauto.
    rewrite purge_app;repeat rewrite undup_equiv_lst;intro x;
    simpl_in; case_in x (support e);tauto.
    rewrite (@purge_l_m_nil (_ ++ (_ ++ _))) in H3
      by (rewrite <- E at 2;intro;simpl_in;intuition);simpl in H3.
    rewrite (@purge_l_m_nil (support f)) in H3
      by (try rewrite <- E;intro;simpl_in;intuition);
    rewrite (@purge_l_m_nil (support e)) in H3
      by (try rewrite <- E;intro;simpl_in;intuition);
      simpl in H3.
    rewrite (@purge_l_m_nil (support f ++ support e)) 
      by (rewrite <- E;intro;simpl_in;intuition);simpl.
    etransitivity;[|eapply eq_ka_impl;try apply ka_s4;apply H3].
    apply eq_plus;auto.
    apply (stronger_ax_proof eqω).
    etransitivity.
    apply eq_smpl_ω_W,eq_W_t.
    apply eq_times;auto;rewrite W_app.
    apply W_equiv_lst;repeat solve_no_dup.
    intro;simpl_in;intuition;try rewrite<-E;simpl_in;tauto.
    rewrite purge_app;repeat rewrite undup_equiv_lst;intro x;
    simpl_in; case_in x (support e);tauto.
    etransitivity.
    apply eq_smpl_ω_W,eq_W_s.
    apply eq_star,W_equiv_lst;repeat solve_no_dup.
    intro;simpl_in;intuition;try rewrite<-E;simpl_in;tauto.
    rewrite<-E at 2;intro;simpl_in;intuition;
    case_ex x (support f);tauto.
  - apply Hyp;repeat split;tauto. 
Qed.

Lemma erase_typed_vers :
  forall e A, str_positive e -> A ⊨ e ->
              eq_perm_ω∨eq_nom_typed∨eq_ω_smpl ⊢
                          W (〈A〉 -- support ⎣e⎦) ⎡⎣e⎦⎤ ≡ e.
Proof.
  induction e;intros A P T;simpl in *;try tauto;auto.
  - rewrite purge_nil;eqr.
    etransitivity;[apply eq_smpl_ω_W,eq_W_id;repeat solve_no_dup;
                   apply disjoint_In;intro;simpl;tauto|].
    apply eq_id;rewrite app_nil_r,undup_equiv_lst;auto.
  - eqr;rewrite (@W_equiv_lst (_ -- _) []);repeat solve_no_dup;auto.
    rewrite undup_equiv_lst,T,purge_l_l;auto.
  - eqr;rewrite (@W_equiv_lst (_ -- _) []);repeat solve_no_dup;auto.
    rewrite undup_equiv_lst,T,purge_l_l;auto.
  - destruct P as (P1&P2);destruct T as (T1 & T2).
    rewrite<- (IHe1 A P1 T1) at 5.
    rewrite<- (IHe2 A P2 T2) at 5.
    eqr;etransitivity;[apply eq_smpl_ω_W,eq_W_p|].
    repeat rewrite W_app.
    pose proof (support_min_type T1) as I1.
    pose proof (support_min_type T2) as I2.
    clear IHe1 IHe2 P1 P2 T1 T2.
    apply eq_plus;apply W_equiv_lst;repeat solve_no_dup;
    intro x;simpl_in;intuition.
    case_in x (support ⎣e2⎦);intuition.
    case_in x (support ⎣e1⎦);intuition.
  - destruct P as (P1&P2);destruct T as (T1 & T2).
    rewrite<- (IHe1 A P1 T1) at 5.
    rewrite<- (IHe2 A P2 T2) at 5.
    eqr;etransitivity;[apply eq_smpl_ω_W,eq_W_t|].
    repeat rewrite W_app.
    pose proof (support_min_type T1) as I1.
    pose proof (support_min_type T2) as I2.
    clear IHe1 IHe2 P1 P2 T1 T2.
    apply eq_times;apply W_equiv_lst;repeat solve_no_dup;
    intro x;simpl_in;intuition.
    case_in x (support ⎣e2⎦);intuition.
    case_in x (support ⎣e1⎦);intuition.
  - simpl;etransitivity;
    [eqr;apply eq_smpl_ω_W,eq_W_s;auto
    |apply eq_star,IHe;auto].
  - destruct T as (I1&T).
    rewrite <-(undup_equiv_lst A a) in I1.
    etransitivity;[|apply eq_ν,IHe;eauto].
    simpl;case_ex a 〈A〉;[tauto|].
    simpl;case inb_spec;intro I2;simpl.
    * rewrite (purge_rm _ I1).
       symmetry;eql;eqr;apply eq_νW2.
       simpl_in;tauto.
       repeat solve_no_dup.
    * rewrite (rm_id I2);eql;eqr.
       etransitivity;[apply W_eq|symmetry];
       eapply eq_ax;[apply (@eqt_νω1 (support ⎣e⎦))|apply (@eqt_νω1 〈A〉)];
       simpl;simpl_beq;simpl;repeat split;
       rewrite (rm_id I2)||rewrite (rm_id I1)||simpl_in;auto.
       apply typed_version_support_type,untyped_erase.
       apply W_type;repeat solve_no_dup;repeat rewrite undup_equiv_lst;
       [eapply ht_convert;
         [|apply typed_version_support_type,untyped_erase]|];
       intro x;simpl_in;intuition.
       pose proof (support_min_type T H) as [->|I];tauto.
  - destruct T as (I1&T);etransitivity;
    [|apply eq_ω,IHe;eauto;now apply str_positive_positive].
    assert (I2: In a (〈A〉 -- support ⎣e⎦)).
    simpl_in;split;auto;intro I2.
    apply support_min_type with (A:=rm a A) in I2;simpl_in;tauto.
    destruct (@W_ω_extract a (〈A〉 -- support ⎣e⎦) ⎡⎣e⎦⎤)
      as (m&ND&L&E);repeat solve_no_dup;auto.
    eqr.
    rewrite E. apply eq_ω,W_equiv_lst;repeat solve_no_dup;auto.
    rewrite<-L,undup_equiv_lst,undup_equiv_lst,rm_purge;auto.
  - etransitivity;[|apply eq_θ,IHe;eauto;
                    now apply str_positive_positive].
    replace (〈A〉 -- p ⧓ support ⎣e⎦)
            with (p⧓ (〈(p ⋆) ⧓ A〉 -- support ⎣e⎦)).
    symmetry;eql;eql;apply eq_perm_ω_W,eq_θ_W.
    clear IHe P T.
    replace 〈_⧓_〉 with (p⋆⧓〈A〉).
    rewrite mact_purge;simpl_perm;auto.
    induction A;simpl;auto.
    rewrite <-IHA.
    case inb_spec; case inb_spec;simpl_in;simpl_perm;
    try tauto.
Qed.
Lemma erase_typed_vers_np :
  forall e A, str_positive e -> no_perm e -> A ⊨ e ->
              eq_nom_typed∨eq_ω_smpl ⊢
                          W (〈A〉 -- support ⎣e⎦) ⎡⎣e⎦⎤ ≡ e.
Proof.
  induction e;intros A P N T;simpl in *;try tauto;auto.
  - rewrite purge_nil;eqr.
    etransitivity;[apply eq_smpl_ω_W,eq_W_id;repeat solve_no_dup;
                   apply disjoint_In;intro;simpl;tauto|].
    apply eq_id;rewrite app_nil_r,undup_equiv_lst;auto.
  - eqr;rewrite (@W_equiv_lst (_ -- _) []);repeat solve_no_dup;auto.
    rewrite undup_equiv_lst,T,purge_l_l;auto.
  - eqr;rewrite (@W_equiv_lst (_ -- _) []);repeat solve_no_dup;auto.
    rewrite undup_equiv_lst,T,purge_l_l;auto.
  - destruct P as (P1&P2);destruct T as (T1 & T2);destruct N as (N1&N2).
    rewrite<- (IHe1 A P1 N1 T1) at 5.
    rewrite<- (IHe2 A P2 N2 T2) at 5.
    apply support_min_type in T1;apply support_min_type in T2;
    clear IHe1 IHe2 P1 P2 N1 N2.
    eqr;etransitivity;[apply eq_smpl_ω_W,eq_W_p|].
    repeat rewrite W_app;apply eq_plus;apply W_equiv_lst;
    repeat solve_no_dup;intro x;pose proof (T1 x) as I1;
    pose proof (T2 x) as I2;simpl_in;intuition.
    case_in x (support ⎣e2⎦);intuition.
    case_in x (support ⎣e1⎦);intuition.
  - destruct P as (P1&P2);destruct T as (T1 & T2);destruct N as (N1&N2).
    rewrite<- (IHe1 A P1 N1 T1) at 5.
    rewrite<- (IHe2 A P2 N2 T2) at 5.
    apply support_min_type in T1;apply support_min_type in T2;
    clear IHe1 IHe2 P1 P2 N1 N2.
    eqr;etransitivity;[apply eq_smpl_ω_W,eq_W_t|].
    repeat rewrite W_app;apply eq_times;apply W_equiv_lst;
    repeat solve_no_dup;intro x;pose proof (T1 x) as I1;
    pose proof (T2 x) as I2;simpl_in;intuition.
    case_in x (support ⎣e2⎦);intuition.
    case_in x (support ⎣e1⎦);intuition.
  - simpl;etransitivity;
    [eqr;apply eq_smpl_ω_W,eq_W_s;auto
    |apply eq_star,IHe;auto].
  - destruct T as (I1&T).
    rewrite <-(undup_equiv_lst A a) in I1.
    etransitivity;[|apply eq_ν,IHe;eauto].
    simpl;case_ex a 〈A〉;[tauto|].
    simpl;case inb_spec;intro I2;simpl.
    * rewrite (purge_rm _ I1).
       symmetry;eql;apply eq_νW2.
       simpl_in;tauto.
       repeat solve_no_dup.
    * rewrite (rm_id I2);eql.
       etransitivity;[apply W_eq|symmetry];
       apply eq_ax;[apply (@eqt_νω1 (support ⎣e⎦))|apply (@eqt_νω1 〈A〉)].
       simpl;simpl_beq;simpl;rewrite (rm_id I2);intuition.
       apply typed_version_support_type,untyped_erase.
       simpl;simpl_beq;simpl;rewrite (rm_id I1);repeat split;auto.
       pose proof (support_min_type T).
       apply W_type;repeat solve_no_dup;
       [eapply ht_convert;
         [|apply typed_version_support_type,untyped_erase]|];
       intro x;simpl_in;intuition.
       destruct (H _ H0) as [->|I];tauto.
  - destruct T as (I1&T);etransitivity;
    [|apply eq_ω,IHe;eauto;now apply str_positive_positive].
    assert (I2: In a (〈A〉 -- support ⎣e⎦)).
    simpl_in;split;auto;intro I2.
    apply support_min_type with (A:=rm a A) in I2;simpl_in;tauto.
    destruct (@W_ω_extract a (〈A〉 -- support ⎣e⎦) ⎡⎣e⎦⎤)
      as (m&ND&L&E);repeat solve_no_dup;auto.
    eqr.
    rewrite E. apply eq_ω,W_equiv_lst;repeat solve_no_dup;auto.
    rewrite<-L,undup_equiv_lst,undup_equiv_lst,rm_purge;auto.
Qed.
Theorem incl_ut_ty :
  forall ax1 ax2 : relation expr,
    Proper (ty'∧ ax2 ==> eq ax1) erase ->
    Proper (ut ∧ ax1 ==> eq ax2) typed_vers ->
    (forall A : list Atom, Proper (ax2 ==> iff)
                                  ( has_type A)) ->
    Proper (ax1 ==> iff) untyped ->
    Proper (ax1 ==> equiv_lst) support ->
    eq_ω_smpl ⇒ ax2 ->
    forall e f, untyped e->untyped f->
                (ax1 ⊢ e ≡ f <-> ax2 ⊢ ⎡e⎤ ≡ ⎡f⎤).
Proof.
  intros ax1 ax2 t1 t2 ty un bs str e f Ue Uf.
  split;intro h.
  - cut ((ut ∧ eq ax1) e f);[|unfold ut;eauto].
    apply eq_typed_vers;auto.
  - rewrite <-(typed_version_erase Ue),<-(typed_version_erase Uf).
    cut ((ty'∧ eq ax2) ⎡e⎤ ⎡f⎤).
    * apply eq_erase;auto.
    * split;auto;exists(support e);split.
       now apply typed_version_support_type.
       rewrite <-(eq_type ty (support e) h).
       now apply typed_version_support_type.
Qed.
Require Import rm0.
Lemma str_positive_eq0 e : str_positive e -> eq0 e = false.
Proof. induction e;simpl;try simpl_bool;firstorder. Qed.
Lemma eq0_erase e : eq0 ⎣e⎦ = eq0 e.
Proof. induction e;simpl;firstorder congruence. Qed.
Theorem incl_ty_ut :
  forall ax1 ax2 : relation expr,
    Proper (ax2 ==> Logic.eq) eq0 ->
    Proper (ty'∧ ax1 ==> eq ax2) erase ->
    Proper (ut ∧ ax2 ==> eq ax1) typed_vers ->
    (forall A : list Atom, Proper (ax1 ==> iff)
                                  ( has_type A)) ->
    Proper (ax2 ==> iff) untyped ->
    Proper (ax2 ==> equiv_lst) support ->
     eq_perm_ω∨eq_nom_typed∨eq_ω_smpl ⇒ ax1 ->
    forall e f, (exists A, A⊨e /\ A⊨f) -> positive e -> positive f ->
                (ax1 ⊢ e ≡ f <-> ax2 ⊢ ⎣e⎦ ≡ ⎣f⎦).
Proof.
  intros ax1 ax2 heq0 t1 t2 ty un bs str e f (A&Te&Tf) Pe Pf.
  apply eq_eq0 in heq0;split;intro h.
  - cut ((ty'∧ eq ax1) e f);[|unfold ty';eauto].
    apply eq_erase;auto.
  - assert (C1 : (exists B, e = ⊥ B) \/ str_positive e)
      by (destruct e;simpl in *;eauto).
    assert (C2 : (exists C, f = ⊥ C) \/ str_positive f)
      by (destruct f;simpl in *;eauto).
    destruct C1 as [(B&->)|Pe'];destruct C2 as [(C&->)|Pf'];
    simpl in *.
    * rewrite <-Te,<-Tf;reflexivity.
    * apply str_positive_eq0 in Pf'.
       apply heq0 in h;rewrite eq0_erase,Pf' in h;discriminate.
    * apply str_positive_eq0 in Pe'.
       apply heq0 in h;rewrite eq0_erase,Pe' in h;discriminate.
    * rewrite<- (stronger_ax_proof str (erase_typed_vers Pe' Te)).
       rewrite<- (stronger_ax_proof str (erase_typed_vers Pf' Tf)).
       rewrite<- (eq_support bs h);apply W_eq.
       cut ((ut ∧ eq ax2) ⎣e⎦ ⎣f⎦).
        apply eq_typed_vers;auto.
        split;auto;split;apply untyped_erase.
Qed.

Theorem incl_ty_ut_np :
  forall ax1 ax2 : relation expr,
    Proper (ax2 ==> Logic.eq) eq0 ->
    Proper (ty'∧ ax1 ==> eq ax2) erase ->
    Proper (ut ∧ ax2 ==> eq ax1) typed_vers ->
    (forall A : list Atom, Proper (ax1 ==> iff)
                                  ( has_type A)) ->
    Proper (ax2 ==> iff) untyped ->
    Proper (ax2 ==> equiv_lst) support ->
    eq_nom_typed∨eq_ω_smpl ⇒ ax1 ->
     forall e f, (exists A, A⊨e /\ A⊨f) -> positive e -> positive f ->
                 no_perm e -> no_perm f ->
                (ax1 ⊢ e ≡ f <-> ax2 ⊢ ⎣e⎦ ≡ ⎣f⎦).
Proof.
  intros ax1 ax2 heq0 t1 t2 ty un bs str e f (A&Te&Tf) Pe Pf Ne Nf.
  apply eq_eq0 in heq0;split;intro h.
  - cut ((ty'∧ eq ax1) e f);[|unfold ty';eauto].
    apply eq_erase;auto.
  - assert (C1 : (exists B, e = ⊥ B) \/ str_positive e)
      by (destruct e;simpl in *;eauto).
    assert (C2 : (exists C, f = ⊥ C) \/ str_positive f)
      by (destruct f;simpl in *;eauto).
    destruct C1 as [(B&->)|Pe'];destruct C2 as [(C&->)|Pf'];
    simpl in *.
    * rewrite <-Te,<-Tf;reflexivity.
    * apply str_positive_eq0 in Pf'.
       apply heq0 in h;rewrite eq0_erase,Pf' in h;discriminate.
    * apply str_positive_eq0 in Pe'.
       apply heq0 in h;rewrite eq0_erase,Pe' in h;discriminate.
    * rewrite<- (stronger_ax_proof str (erase_typed_vers_np Pe' Ne Te)).
       rewrite<- (stronger_ax_proof str (erase_typed_vers_np Pf' Nf Tf)).
       rewrite<- (eq_support bs h);apply W_eq.
       cut ((ut ∧ eq ax2) ⎣e⎦ ⎣f⎦).
        apply eq_typed_vers;auto.
        split;auto;split;apply untyped_erase.
Qed.