(** printing ψ %\ensuremath{\psi}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Import proofs.


Variable φ : Atom -> Atom.
Variable ψ : Atom -> Atom.
Variable ψ_φ_id : forall x, ψ (φ x) = x.

Lemma φ_inj : forall x y, φ x = φ y -> x = y.
Proof. now intros;rewrite<-(ψ_φ_id x),H,ψ_φ_id. Qed.

Definition X x := exists y, x = φ y.
Definition in_X : Atom -> bool := fun x => beq_atom x (φ (ψ x)).

Lemma in_X_spec x : reflect (X x) (in_X x).
Proof.
  unfold X,in_X;case_eq_at x (φ (ψ x));
  try rewrite ψ_φ_id in *;simpl_beq; [apply ReflectT|apply ReflectF].
  eexists;eauto.
  now intros (y&->);rewrite ψ_φ_id in *;apply h.
Qed.
Lemma in_X_X x : (X x) <-> (in_X x = true).
Proof. apply (reflect_iff _ _ (in_X_spec x)). Qed.

Lemma φ_X : forall x, X (φ x).
Proof. now intro;eexists;eauto. Qed.

Lemma φ_in_X : forall x, in_X (φ x) = true.
Proof. now intro;rewrite<-in_X_X;apply φ_X. Qed.


Lemma φ_inv x y : x = y <-> φ x = φ y.
Proof. intuition subst;auto using φ_inj. Qed.
      
Definition restr_perm θ :=
  forallb (fun p => match p with (a,b) => in_X a && in_X b end) θ.
Lemma restr_perm_act θ a : restr_perm θ = true -> X a -> X (θ⋈a).
Proof.
  revert a;induction θ;try now intuition.
  destruct a as (a,b);intros ? h1 (x,->).
  simpl in *;repeat rewrite andb_true_iff in h1;
  destruct h1 as ((h1&h2)&h3).
  apply in_X_X in h1 as (y&->);apply in_X_X in h2 as (z&->).
  case_eq_at (θ⋈φ x) (φ y);simpl_beq.
  * eexists; eauto.
  * case_eq_at (θ⋈φ x) (φ z);simpl_beq.
  - eexists; eauto.
  - apply IHθ;[|eexists];eauto.
Qed.
Definition restr_list := forallb in_X.
Lemma restr_perm_mact θ l :
  restr_perm θ=true -> restr_list l=true -> restr_list (θ⧓l)=true.
Proof.
  induction l;intuition;simpl in *.
  rewrite andb_true_iff in *;intuition.
  rewrite <-in_X_X in *; now apply restr_perm_act.
Qed.

Fixpoint restr_expr e :=
  match e with
  | id A | bot A => restr_list A
  | 0 | 1 | var _ | vat _ => true
  | e + f | e · f => restr_expr e && restr_expr f
  | e ^* => restr_expr e
  | ν a e | ω a e => in_X a && restr_expr e
  | ‹ p › e => restr_perm p && restr_expr e
  end.

Definition is_restr e := restr_expr e = true.

Definition map_pair (f : Atom -> Atom) :=
  map (fun x => (f (fst x),f (snd x))).
Lemma map_pair_inv f p : (map_pair f p)⋆ = map_pair f (p⋆).
Proof. unfold map_pair;induction p;simpl;try rewrite map_app,IHp;simpl;
       firstorder.
Qed.
Lemma map_φ_act p a : (map_pair φ p) ⋈ φ a = φ (p ⋈ a).
Proof.
    induction p;simpl;auto.
    rewrite IHp.
    destruct a0.
    case_eq_at (p⋈a) a0;simpl;repeat simpl_beq;auto.
    rewrite φ_inv in h;simpl_beq.
    case_eq_at (p⋈a) a1;simpl;repeat simpl_beq;auto.
    rewrite φ_inv in h0;simpl_beq;auto.
Qed.
Lemma restr_perm_map_φ p :
  restr_perm p = true -> p = map_pair φ (map_pair ψ p).
Proof.
  intros. unfold restr_perm in H.
  induction p;simpl in *;repeat simpl_bool;intuition.
  f_equal;auto.
  destruct a as (x,y);simpl_bool;rewrite <-in_X_X in *.
  simpl;destruct H as (?&->);destruct h0 as (?&->).
  now repeat rewrite ψ_φ_id.
Qed.

Lemma map_ψ_map_φ_act_φ p a : 
  (map_pair ψ (map_pair φ p)) ⋈ a = ψ ((map_pair φ p) ⋈ (φ a)).
Proof.
  rewrite map_φ_act,ψ_φ_id.
  unfold map_pair;rewrite map_map.
  rewrite<- (map_id p) at 2.
  f_equal;apply map_ext.
  now intros (x,y);simpl;repeat rewrite ψ_φ_id.
Qed.
Lemma map_ψ_act_φ p a :
  restr_perm p = true -> (map_pair ψ p) ⋈ a = ψ (p ⋈ (φ a)).
Proof. now intro h;rewrite (restr_perm_map_φ h),map_ψ_map_φ_act_φ. Qed.
Lemma map_φ_mact p A : map φ (p ⧓ A) = (map_pair φ p) ⧓ map φ A.
Proof. induction A;simpl;auto;rewrite IHA,map_φ_act;auto. Qed.

Lemma map_φ_rm a A : map φ (rm a A) = rm (φ a) (map φ A).
Proof. induction A;simpl;auto; case_eq_at a a0;simpl;
       try simpl_beq;auto;now rewrite φ_inv in h;simpl_beq;rewrite IHA.
Qed.

Definition perm_fun_lst f l := transfer_perm 〈l〉 (map f 〈l〉).
Lemma perm_fun_lst_mact f : forall l,
    NoDup (map f 〈l〉) ->
    (perm_fun_lst f l) ⧓ l = map f l.
Proof.
  cut (forall l m, 
          NoDup (map f 〈m〉) ->
          l⊆m -> perm_fun_lst f m ⧓ l = map f l).
  now intros h nd l;rewrite h.
  induction l;intuition.
  simpl;rewrite IHl;[|auto |now intros ? ?;apply H0;right].
  f_equal.
  assert (perm_fun_lst f m ⧓ 〈m〉 = map f 〈m〉).
  - clear H0 l a IHl.
    unfold perm_fun_lst;apply is_perm_no_dup_length;repeat solve_no_dup;
    induction m;simpl in *;intuition;repeat solve_no_dup.
    revert H;case inb_spec;intros;simpl;rewrite IHm;auto.
    now simpl in *; rewrite NoDup_cons_iff in H.
  - cut (forall a, In a 〈m〉 -> perm_fun_lst f m ⋈ a = f a).
    intro h;apply h;simpl_in;apply H0;now left.
    clear a l IHl H0;induction 〈m〉;firstorder.
    now simpl in H1;inversion H1;rewrite <-H0.
    apply IHl;auto.
    now simpl in *; rewrite NoDup_cons_iff in H.
    now simpl in H1;inversion H1.
Qed.

Fixpoint apply_fun f e :=
  match e with
  | id A => id (map f A)
  | ⊥ A => ⊥ (map f A)
  | var x => (perm_fun_lst f (τ x))•var x
  | vat a => vat (f a)
  | e1 + e2 => apply_fun f e1 + apply_fun f e2
  | e1 · e2 => apply_fun f e1 · apply_fun f e2
  | e^* => (apply_fun f e)^*
  | ν a e => ν (f a) (apply_fun f e)
  | ω a e => ω (f a) (apply_fun f e)
  | ‹ p › e => ‹ map_pair f p › apply_fun f e   
  | _ => e
  end.

Lemma apply_φ_restr e : is_restr (apply_fun φ e).
Proof.
  unfold is_restr;induction e;simpl;repeat rewrite andb_true_iff in *;
  intuition;try apply φ_in_X.
  * induction l;simpl;try rewrite andb_true_iff;intuition;apply φ_in_X.
  * induction l;simpl;try rewrite andb_true_iff;intuition;apply φ_in_X.
  * unfold map_pair;apply forallb_forall;intros (?,?);
    rewrite in_map_iff;
    intros ((?,?),?);simpl in *;rewrite andb_true_iff.
    destruct H as (H,_);inversion H;split;apply φ_in_X.
Qed.

Lemma map_φ_inj l m : l ≈ m <-> map φ l ≈ map φ m.
Proof.
  split.
  * intros H x;repeat rewrite in_map_iff;simpl;split;intros (?&<-&h);
    rewrite H in h || rewrite <- H in h;firstorder.
  * unfold equiv_lst; intros H x;
    repeat setoid_rewrite in_map_iff in H;simpl;intuition.
    cut (exists x0, φ x0 = φ x /\ In x0 l).
    rewrite H;intros (y&E&I);apply φ_inj in E as ->;auto.
    eexists;eauto.
    cut (exists x0, φ x0 = φ x /\ In x0 m).
    rewrite <-H;intros (y&E&I);apply φ_inj in E as ->;auto.
    eexists;eauto.
Qed.
Lemma NoDup_map_φ l : NoDup (map φ l) <-> NoDup l.
Proof.
  induction l;intuition;simpl in *;rewrite NoDup_cons_iff in *;
  rewrite NoDup_cons_iff in H1;rewrite in_map_iff in *;firstorder.
  now intros (?&h1&?);apply φ_inj in h1 as ->.
Qed.
Lemma apply_φ_no_perm e : no_perm e <-> no_perm (apply_fun φ e).
Proof. induction e;simpl;intuition. Qed.
Lemma apply_φ_multi e : multi e <-> multi (apply_fun φ e).
Proof. induction e;simpl;intuition. Qed.
Lemma apply_φ_single e : single e <-> single (apply_fun φ e).
Proof. induction e;simpl;intuition. Qed.
Lemma apply_φ_untyped e : untyped e <-> untyped (apply_fun φ e).
Proof. induction e;simpl;intuition. Qed.
Lemma apply_φ_has_type e A : A ⊨ e <-> map φ A ⊨ (apply_fun φ e).
Proof.
  revert A;induction e;simpl;intro;try (now firstorder).
  * now rewrite<- map_φ_inj.
  * now rewrite<- map_φ_inj.
  * rewrite apply_perm_var_τ,perm_fun_lst_mact,<- map_φ_inj;try tauto.
    rewrite NoDup_map_φ;solve_no_dup.
  * replace [φ a] with (map φ [a]) by (simpl;auto);
    now rewrite<- map_φ_inj.
  * rewrite in_map_iff;rewrite IHe;simpl;firstorder.
    intros (x&E&I);apply φ_inj in E as ->;tauto.
  * rewrite <- map_φ_rm,in_map_iff;rewrite IHe;simpl;firstorder.
    apply φ_inj in H as ->;tauto.
  * rewrite map_pair_inv;rewrite IHe;split;apply ht_convert;
    rewrite map_φ_mact;auto.
Qed.

Lemma apply_φ_str_positive e :
  str_positive e <-> str_positive (apply_fun φ e).
Proof. induction e;simpl;intuition. Qed.
Lemma apply_φ_positive e :
  positive e <-> positive (apply_fun φ e).
Proof. induction e;simpl;intuition;
       try rewrite <-apply_φ_str_positive in *;auto.
Qed.
Lemma elements_map_pair f p :
  elements (map_pair f p) ≈ map f (elements p).
Proof. induction p;simpl;intuition;rewrite IHp;auto. Qed.
Lemma equiv_perm_map_φ p q : p ≌ q <-> map_pair φ p ≌ map_pair φ q.
Proof.
  split;intro;rewrite mact_elements_equiv_perm.
  setoid_rewrite in_app_iff;setoid_rewrite elements_map_pair;
  setoid_rewrite in_map_iff.
  intros x [(y&<-&I)|(y&<-&I)];
    repeat rewrite map_φ_act;rewrite<- φ_inv;auto.
  setoid_rewrite in_app_iff; intros x [I|I];
  apply φ_inv;repeat rewrite <- map_φ_act;auto.
Qed.

Lemma apply_φ_eq_expr e f :
  (e == f) <-> (apply_fun φ e == apply_fun φ f).
Proof.
  revert f;induction e;intro f;destruct f;simpl;try tauto.
  * apply map_φ_inj.
  * apply map_φ_inj.
  * destruct l as ((?,?),?);destruct l0 as ((?,?),?);unfold eq_letter;
    unfold τ;simpl;split;intro H;inversion H; f_equal;auto.
    rewrite perm_fun_lst_mact, perm_fun_lst_mact in H2;
    try (apply NoDup_map_φ;solve_no_dup).
    clear n n0 e e0 H H1.
    revert l0 H2;induction l;intros [| x m ] h;
    simpl in *; discriminate || auto.
    inversion h;rewrite <- φ_inv in H0; rewrite H0,(IHl m);auto.
  * apply φ_inv.
  * firstorder.
  * firstorder.
  * firstorder.
  * rewrite φ_inv;firstorder.
  * rewrite φ_inv;firstorder.
  * rewrite equiv_perm_map_φ;firstorder.
Qed.

Lemma apply_φ_support e : support (apply_fun φ e) ≈ map φ (support e).
Proof.
  induction e;simpl;auto.
  * now destruct l as ((?,?),?);unfold τ;simpl;
    rewrite perm_fun_lst_mact;try (apply NoDup_map_φ;solve_no_dup).
  * now rewrite IHe1,IHe2,map_app.
  * now rewrite IHe1,IHe2,map_app.
  * now rewrite IHe, map_φ_rm.
  * now rewrite map_φ_mact,IHe.
Qed.

Lemma apply_φ_stage e : stage (apply_fun φ e) ≈ map φ (stage e).
Proof.
  induction e;simpl;auto.
  * now destruct l as ((?,?),?);unfold τ;simpl;
    rewrite perm_fun_lst_mact;try (apply NoDup_map_φ;solve_no_dup).
  * now rewrite IHe1,IHe2,map_app.
  * now rewrite IHe1,IHe2,map_app.
  * now rewrite IHe, map_φ_rm.
  * now rewrite IHe.
  * now rewrite map_φ_mact,IHe.
Qed.

Lemma apply_φ_fresh e a : a # e <-> φ a # (apply_fun φ e).
Proof.
  revert a;induction e;simpl;intro;try (now firstorder).
  * rewrite apply_perm_var_τ,perm_fun_lst_mact,in_map_iff;firstorder.
    now intros (x&E&I);apply H;apply φ_inv in E as ->.
    (apply NoDup_map_φ;solve_no_dup).
  * now rewrite φ_inv.
  * now rewrite φ_inv,IHe.
  * now rewrite map_pair_inv,map_φ_act,IHe.
Qed.

Lemma app_length_eq A :
  forall (l m n o: list A),
    l++m=n++o /\ length l = length n -> l = n /\ m = o.
Proof.
  induction l;intros m n o (E&L);destruct n;simpl in *;
  discriminate || auto.
  inversion L; clear L;inversion E;clear E.
  rewrite <- H1 in *;clear H1.
  pose proof (IHl m n o) as (h1 & h2);auto.
  split;[f_equal|];auto.
Qed.
Lemma φ_lst_act l a : In a l -> perm_fun_lst φ l ⋈ a = φ a.
Proof.
  intro h;apply in_split in h as (?&?&?).
  pose proof (@app_length_eq _ ((perm_fun_lst φ l) ⧓ x)
                             (perm_fun_lst φ l ⧓ (a::x0))
                             (map φ x)  (map φ (a::x0))) as (h1 & h2);
    auto.
  - split.
    * rewrite<-map_app,<-H, <-perm_fun_lst_mact,H;auto;
      try (apply NoDup_map_φ;solve_no_dup).
      now unfold mact;rewrite map_app.
    * now unfold mact;repeat rewrite map_length.
  - simpl in h2;inversion h2;auto.
Qed.

Fixpoint all_atoms e :=
  match e with
  | 0 | 1 => []
  | bot A | id A => A
  | var x => τ x
  | vat a => [a]
  | e + f | e · f => (all_atoms e) ++ (all_atoms f)
  | e ^* => (all_atoms e)
  | ‹ p › e => (elements p)++(all_atoms e)
  | ν a e | ω a e => a::(all_atoms e)
  end.
Lemma φ_lst_incl_mact l m : l⊆m -> (perm_fun_lst φ m) ⧓ l = map φ l.
Proof.
  induction l;simpl;intro;auto.
  rewrite φ_lst_act,IHl;auto.
  intros ? ?;apply H;auto.
Qed.
Lemma apply_φ_apply_perm e l : (all_atoms e) ⊆ l -> no_perm e ->
                apply_fun φ e == (perm_fun_lst φ l) • e.
Proof.
  induction e;simpl;intuition.
  * rewrite φ_lst_incl_mact;auto.
  * rewrite φ_lst_incl_mact;auto.
  * simpl;destruct l0 as ((?,?),?);
    unfold τ,eq_letter in *;simpl in *;rewrite perm_fun_lst_mact;
    f_equal;try (apply NoDup_map_φ;solve_no_dup).
    now symmetry;apply φ_lst_incl_mact.
  * symmetry;apply φ_lst_act;now apply H;left.
  * apply IHe1;auto;intros ? ?;apply H;rewrite In_app_iff;tauto.
  * apply IHe2;auto;intros ? ?;apply H;rewrite In_app_iff;tauto.
  * apply IHe1;auto;intros ? ?;apply H;rewrite In_app_iff;tauto.
  * apply IHe2;auto;intros ? ?;apply H;rewrite In_app_iff;tauto.
  * apply IHe;auto;intros ? ?;apply H;simpl;tauto.
  * rewrite φ_lst_act;auto.
  * apply IHe;auto;intros ? ?;apply H;simpl;tauto.
  * rewrite φ_lst_act;auto.
Qed.
Lemma apply_perm_app p q e : (p++q)•e == p • (q • e).
Proof.
  revert p q;induction e;simpl;intuition.
  * now rewrite mact_comp_app.
  * now rewrite mact_comp_app.
  * now rewrite apply_perm_var_app.
  * now rewrite act_comp_app.
  * now rewrite act_comp_app.
  * now rewrite act_comp_app.
  * now rewrite <-IHe,app_nil_r,app_ass.
Qed.


Lemma aux e l : (all_atoms e) ⊆ l -> no_perm e ->
                FFalse ⊢ apply_fun φ e ≡ (perm_fun_lst φ l) • e.
Proof. now intros;apply eq_expr_eq,apply_φ_apply_perm. Qed.
Lemma aux2 e l : (all_atoms e) ⊆ l -> no_perm e ->
                 FFalse ⊢ e ≡(perm_fun_lst φ l)⋆ • apply_fun φ e.
Proof. intros; apply eq_expr_eq;rewrite apply_φ_apply_perm;eauto.
       now rewrite <-apply_perm_app,perm_pinv_p_eq_nil,apply_perm_nil. 
Qed.


Lemma apply_φ_eq_ax (ax : relation expr) :
  (forall e f, ax e f -> ax ⊢  apply_fun φ e ≡ apply_fun φ f) ->
  forall e f, ax ⊢ e ≡ f -> ax ⊢ apply_fun φ e ≡ apply_fun φ f.
Proof.
  intros hyp e f E;induction E;auto;
  try (now apply eq_expr_eq; rewrite <-apply_φ_eq_expr).
  * eauto.
  * now simpl;rewrite IHE1,IHE2.
  * now simpl;rewrite IHE1,IHE2.
  * now simpl;rewrite IHE.
  * now simpl;rewrite IHE.
  * now simpl;apply equiv_perm_map_φ in H;rewrite IHE,H.
  * now simpl;rewrite IHE.
  * destruct H;simpl;auto.
  * destruct H;simpl in *;eauto.
Qed.

Definition restr_ax ax := fun e f =>
                            ax e f /\ is_restr e /\ is_restr f.

Lemma restr_ax_eq_ax_eq ax e f :
  restr_ax ax ⊢ e ≡ f -> ax ⊢ e ≡ f.
Proof. intro; apply (@stronger_ax_proof (restr_ax ax));
       unfold restr_ax in *;firstorder.
Qed.

Definition ex_ax_eq ax : relation expr :=
  fun e f => exists g, ax e g /\ (g==f) /\ (is_restr g <-> is_restr f).
Lemma proper_ex_ax_eq ax ψ :
  Proper (ax ==> ax) ψ ->
  Proper (ax ==> ex_ax_eq ax) ψ.
Proof. intros h e f E;apply h in E;exists (ψ f);intuition. Qed.
Lemma ax_eq_restr_ax_eq ax :
  Proper (ax ==> ex_ax_eq ax) (apply_fun φ) ->
  Proper (eq ax ==> eq (restr_ax ax)) (apply_fun φ).
Proof.
  intros h e f E;induction E;simpl;auto.
  * etransitivity;eauto.
  * now apply eq_bot; rewrite<- map_φ_inj.
  * now apply eq_id; rewrite<- map_φ_inj.
  * now apply eq_var;destruct x as ((?,?),?);destruct y as ((?,?),?);
    unfold τ,eq_letter in *;simpl in *;inversion H.
  * apply equiv_perm_map_φ in H as ->;eauto.
  * destruct H;simpl;auto.
  * destruct H;simpl in *;eauto.
  * apply h in H as (g&h1&h2&h3).
    transitivity g.
    apply eq_ax;repeat split; auto;try rewrite h3;apply apply_φ_restr.
    now apply eq_expr_eq.
Qed.

Lemma zero_ka_u_apply_φ : Proper (zero_ka_u ==> zero_ka_u) (apply_fun φ).
Proof. intros e f E;destruct E;simpl in *;auto;
       apply apply_φ_untyped in H;auto.
Qed.

Lemma zero_ka_t_apply_φ : Proper (zero_ka_t ==> zero_ka_t) (apply_fun φ).
Proof. intros e f E;destruct E;simpl in *;auto;
       apply apply_φ_has_type in H;auto.
Qed.

Lemma zero_perm_u_apply_φ :
  Proper (zero_perm_u ==> zero_perm_u) (apply_fun φ).
Proof. intros e f E;destruct E;simpl in *;auto. Qed.
    
Lemma zero_perm_t_apply_φ :
  Proper (zero_perm_t ==> zero_perm_t) (apply_fun φ).
Proof. intros e f E;destruct E;simpl in *;
       rewrite map_φ_inj,map_φ_mact in H;eauto.
Qed.

Lemma zero_nom_u_apply_φ : Proper (zero_nom_u ==> zero_nom_u) (apply_fun φ).
Proof. intros e f E;destruct E;simpl in *;apply equ_ν0. Qed.
       
Lemma zero_nom_t_apply_φ : Proper (zero_nom_t ==> zero_nom_t) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;auto.
  * apply eq_ω_bot;intro h;apply in_map_iff in h as (y&h1&h2);
    apply φ_inj in h1 as ->;tauto.
  * rewrite map_φ_rm;apply eqt_ν0.
    apply in_map_iff;exists a;auto.
Qed.
    
Lemma eq_perm_apply_φ : Proper (eq_perm ==> eq_perm) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;eauto.
  * rewrite <-map_φ_act; apply eq_θ_ν.
  * unfold map_pair;rewrite map_app; auto.
Qed.

Lemma eq_perm_ω_apply_φ : Proper (eq_perm_ω ==> eq_perm_ω) (apply_fun φ).
Proof. intros e f E;destruct E;simpl in *;rewrite <-map_φ_act;auto. Qed.

Lemma eq_perm_elim_apply_φ :
  Proper (eq_perm_elim ==> ex_ax_eq eq_perm_elim) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;eauto.
  * exists 1;simpl;intuition.
  * rewrite <-map_φ_act; auto;eexists;firstorder.
  * exists
      (var (apply_perm_var (map_pair φ p) (apply_perm_var (perm_fun_lst φ (τ a)) a))).
    destruct a as ((?,?),?);unfold τ,eq_letter in *;simpl in *.
    repeat  rewrite perm_fun_lst_mact|| rewrite map_φ_mact.
    repeat split.
    apply eq_θ_var.
    unfold τ,eq_letter in *;simpl in *.
    now rewrite perm_fun_lst_mact, perm_fun_lst_mact, map_φ_mact;
      try (apply NoDup_map_φ;solve_no_dup).
  * exists (id (map φ B)); intuition.
    now apply eq_θ_id; rewrite <-map_φ_mact,<-map_φ_inj.
  * exists (apply_fun φ e);intuition.
Qed.
    
Lemma eq_ν_smpl_apply_φ : Proper (eq_ν_smpl ==> eq_ν_smpl) (apply_fun φ).
Proof. intros e f E;destruct E;simpl in *;eauto. Qed.
Lemma eq_ω_smpl_apply_φ : Proper (eq_ω_smpl ==> eq_ω_smpl) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;eauto.
  apply eq_ω_id;rewrite in_map_iff;intros (x&h1&h2);
  now apply φ_inj in h1 as ->.
Qed.

Lemma eq_nom_untyped_p_apply_φ :
  Proper (eq_nom_untyped_p ==> eq_nom_untyped_p) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;eauto.
  now apply equ_θν';rewrite <-apply_φ_fresh.
Qed.
Lemma apply_φ_apply_perm_p p e :
  apply_fun φ (p • e) == (map_pair φ p) • (apply_fun φ e).
Proof.
  revert p;induction e;intro q;simpl;f_equal;auto.
  * now rewrite map_φ_mact.
  * now rewrite map_φ_mact.
  * destruct l as ((?,?),?);unfold τ,eq_letter;simpl.
    now rewrite perm_fun_lst_mact,perm_fun_lst_mact,map_φ_mact;
      try (apply NoDup_map_φ;solve_no_dup).
  * now rewrite map_φ_act.
  * now rewrite map_φ_act.
  * now rewrite map_φ_act.
  * now unfold map_pair in *;rewrite<- map_app.
Qed.
(* Inductive eq_nom_untyped_np' : relation expr := *)
(*   equ'_θν a b e f : *)
(*     a # e -> (f == [a, b] • e) -> eq_nom_untyped_np' (ν b e) (ν a f). *)

Lemma restr_perm_is_restr :
  (forall p e, restr_perm p = true -> is_restr (p•apply_fun φ e)).
Proof.
  intros p e;unfold is_restr;revert p;induction e;
  intuition;simpl.
  apply restr_perm_mact;auto.
  clear p H;induction l;simpl;auto with bool.
  rewrite φ_in_X,IHl;auto with bool.
  apply restr_perm_mact;auto.
  clear p H;induction l;simpl;auto with bool.
  rewrite φ_in_X,IHl;auto with bool.
  rewrite IHe1,IHe2;auto with bool.
  rewrite IHe1,IHe2;auto with bool.
  rewrite andb_true_iff;split.
  rewrite <-in_X_X; apply restr_perm_act; auto;unfold X;apply φ_X.
  rewrite IHe;auto.
  rewrite andb_true_iff;split.
  rewrite <-in_X_X; apply restr_perm_act; auto;unfold X;apply φ_X.
  rewrite IHe;auto.
  rewrite IHe;auto.
  assert (restr_perm (map_pair φ p)=true).
  clear e IHe p0 H;induction p;simpl;auto.
  repeat (rewrite andb_true_iff;split);
    auto;unfold X;rewrite φ_in_X;auto.
  unfold restr_perm in *;rewrite forallb_app,H,H0;auto.
Qed.
Lemma eq_nom_untyped_np_apply_φ :
  Proper (eq_nom_untyped_np ==> ex_ax_eq eq_nom_untyped_np) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;eauto.
  exists (ν (φ a)([φ a,φ b]•(apply_fun φ e)));split.
  apply equ_θν.
  now rewrite <-apply_φ_fresh.
  split.
  simpl;intuition.
  now rewrite apply_φ_apply_perm_p.
  unfold is_restr;simpl.
  pose proof (φ_in_X a) as I;unfold X in I;rewrite I.
  rewrite apply_φ_restr.
  rewrite restr_perm_is_restr;intuition auto with bool.
  simpl;repeat  rewrite φ_in_X; auto with bool.
Qed.
Lemma eq_nom_untyped_apply_φ :
  Proper (eq_nom_untyped ==> eq_nom_untyped) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;rewrite apply_φ_fresh in H;auto.
Qed.
Lemma eq_nom_typed_p_apply_φ :
  Proper (eq_nom_typed_p ==> eq_nom_typed_p) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;intuition.
  eapply eqt_θν'. apply apply_φ_has_type in H2;simpl in H2.
  simpl;split;eauto.
  rewrite in_map_iff;intros (x&h1&h2);apply φ_inj in h1 as ->;tauto.
  rewrite in_map_iff;intros (x&h1&h2);apply φ_inj in h1 as ->;tauto.
Qed.

Lemma eq_nom_typed_np_apply_φ :
  Proper (eq_nom_typed_np ==> ex_ax_eq eq_nom_typed_np) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;intuition.
  exists (ν (φ b) ([φ b, φ a] • apply_fun φ e));split.
  eapply eqt_θν. 
  apply apply_φ_has_type in H2;simpl in H2.
  simpl;split;eauto.
  rewrite in_map_iff;intros (x&h1&h2);apply φ_inj in h1 as ->;tauto.
  rewrite in_map_iff;intros (x&h1&h2);apply φ_inj in h1 as ->;tauto.
  split.
  replace [φ b,φ a] with (map_pair φ [b,a]);auto.
  simpl;split;auto. symmetry. apply apply_φ_apply_perm_p.
  unfold is_restr;simpl.
  rewrite apply_φ_restr.
  rewrite restr_perm_is_restr;intuition auto with bool.
  simpl;repeat  rewrite φ_in_X; auto with bool.
Qed.
  
Lemma eq_nom_typed_apply_φ :
  Proper (eq_nom_typed ==> eq_nom_typed) (apply_fun φ).
Proof.
  intros e f E;destruct E;simpl in *;intuition.
  * assert (I:~ In (φ a) (map φ A)).
    rewrite in_map_iff;intros (x&h1&h2);apply φ_inj in h1 as ->;tauto.
    simpl_beq; eapply eqt_νω1;simpl;simpl_beq;intuition;eauto.
    now rewrite<-map_φ_rm,<-apply_φ_has_type.
  * assert (E:φ a <> φ b) by (rewrite <-φ_inv;auto);eauto.
Qed.

Lemma is_restr_stable ax e f :
  no_perm e -> no_perm f ->
  Proper (ax ==> iff) no_perm ->
  restr_ax ax ⊢ e ≡ f -> is_restr e <-> is_restr f.
Proof.
  intros n1 n2 npax E.
  assert (npra:Proper (eq (restr_ax ax) ==> iff) no_perm).
  clear e f n1 n2 E;intros e f E;induction E;firstorder.
  destruct H;firstorder.
  destruct H;firstorder.
  destruct H;firstorder.

  induction E;unfold is_restr in *;simpl;try unfold restr_list;
  try tauto.
  * apply npra in E1;firstorder.
  * repeat rewrite forallb_forall;firstorder.
  * repeat rewrite forallb_forall;firstorder.
  * repeat rewrite andb_true_iff;firstorder.
  * repeat rewrite andb_true_iff;firstorder.
  * repeat rewrite andb_true_iff;firstorder.
  * repeat rewrite andb_true_iff;firstorder.
  * repeat rewrite andb_true_iff;firstorder.
  * destruct H;simpl;repeat rewrite andb_true_iff;firstorder.
  * destruct H;simpl in *;repeat rewrite andb_true_iff in *;
    firstorder.
  * unfold restr_ax,is_restr in H;tauto.
Qed.

Lemma apply_perm_eq ax e f p:
  (forall p, Proper (ax ==> eq ax) (apply_perm p))
  -> ax ⊢ e ≡ f ->  ax ⊢ p•e≡p•f.
Proof.
  intros h E;revert p;induction E;simpl;firstorder.
  * eauto.
  * now rewrite H.
  * now rewrite H.
  * apply eq_var;destruct x as ((?,?),?);
    destruct y as ((?,?),?);unfold eq_letter,τ in *; simpl in *.
    now inversion H.
  * apply eq_θ;auto.
    etransitivity;[apply eq_expr_eq;apply apply_perm_equiv_perm;
                   [rewrite H|];reflexivity|];firstorder.
  * destruct H;simpl;eauto.
  * destruct H;simpl in *;eauto.
Qed.

Lemma apply_fun_comp f g e :
  (forall l, NoDup l -> NoDup (map g l)) ->
  (forall l, NoDup l -> NoDup (map f (map g l))) ->
  apply_fun f (apply_fun g e) == apply_fun (f ∘ g) e.
Proof.
  intros hyp1 hyp2;induction e;simpl;auto.
  * rewrite map_map;auto.
  * rewrite map_map;auto.
  * pose proof (τ_spec l) as h.
    destruct l as ((?,?),?);unfold τ,eq_letter in *;simpl in *;f_equal.
    rewrite perm_fun_lst_mact, perm_fun_lst_mact,
    perm_fun_lst_mact,map_map;auto;rewrite<-NoDup_undup_eq;auto;
    [unfold Basics.compose;rewrite<-map_map| |];
    try apply hyp2;try apply hyp1;try solve_no_dup;auto.
    rewrite perm_fun_lst_mact; apply hyp1 || apply hyp2;auto;
    solve_no_dup.
  * split;auto;unfold map_pair;rewrite map_map;auto.
Qed.

Lemma apply_fun_id e : apply_fun (fun x => x) e == e.
Proof.
  induction e;simpl;auto;try rewrite map_id;auto.
  * destruct l as ((?,?),?);unfold τ,eq_letter;simpl.
    rewrite perm_fun_lst_mact;rewrite map_id;auto;solve_no_dup.
  * split;auto.
    unfold map_pair.
    erewrite map_ext.
    rewrite map_id;auto.
    intros (?,?);firstorder.
Qed.
Lemma apply_fun_ext f g e : (forall x, f x = g x) ->
                            apply_fun f e = apply_fun g e.
Proof.
  intro h;induction e;simpl;
  try rewrite (map_ext f g h) || rewrite h; f_equal;try firstorder.
  * now unfold perm_fun_lst;rewrite (map_ext f g h).
  * auto;unfold map_pair;erewrite map_ext;eauto.
    now intros (?,?);simpl;repeat rewrite h.
Qed.
Lemma apply_ψ_φ_id e : apply_fun ψ (apply_fun φ e) == e.
Proof.
  rewrite apply_fun_comp.
  - erewrite apply_fun_ext.
    -- apply apply_fun_id.
    -- intros;simpl;apply ψ_φ_id.
  - now intros;rewrite NoDup_map_φ.
  - intros;rewrite map_map;erewrite map_ext,map_id;auto.
    intros;simpl;apply ψ_φ_id.
Qed.
Lemma elements_In :
  forall p x,
    In x (elements p) <-> exists y, In y p /\ (x=fst y \/ x = snd y).
Proof.
  induction p;intros x;simpl.
  - firstorder.
  - destruct a as (b,c);simpl;rewrite IHp;clear IHp;split;
    [intros [-> | [->|((d,e)&h1&[->| ->])]]
    |intros ((d,e)&[h1|h1]&[-> | ->])];simpl;
    try (eexists;eauto)
    ||(inversion h1;auto)||right;right;eexists;eauto.
Qed.
Lemma equiv_perm_map_ψ p q :
  restr_perm p = true ->restr_perm q = true ->
  p ≌ q <-> map_pair ψ p ≌ map_pair ψ q.
Proof.
  intros hp hq;simpl in *.
  split;intro;rewrite mact_elements_equiv_perm.
  setoid_rewrite in_app_iff;setoid_rewrite elements_map_pair;
  setoid_rewrite in_map_iff.
  intros x [(y&<-&I)|(y&<-&I)].
  repeat rewrite  map_ψ_act_φ;auto.
  now rewrite H.
  repeat rewrite  map_ψ_act_φ;auto.
  now rewrite H.
  intros x I.
  
  assert (x = φ (ψ x)).
  apply in_app_iff in I as [I|I].
  unfold restr_perm in hp;rewrite forallb_forall in hp.
  apply elements_In in I as ((x1&x2)&h1&h2);apply hp in h1.
  simpl_bool;rewrite <- in_X_X in *;simpl in *.
  destruct h1 as (y1&->);destruct h as (y2&->).
  now destruct h2 as [->| ->];rewrite ψ_φ_id.
  unfold restr_perm in hq;rewrite forallb_forall in hq.
  apply elements_In in I as ((x1&x2)&h1&h2);apply hq in h1.
  simpl_bool;rewrite <- in_X_X in *;simpl in *.
  destruct h1 as (y1&->);destruct h as (y2&->).
  now destruct h2 as [->| ->];rewrite ψ_φ_id.
  now rewrite H0,(restr_perm_map_φ hp), (map_φ_act),
  (restr_perm_map_φ hq), (map_φ_act),H.
Qed.  


Lemma apply_fun_morph : Proper (restr_ax eq_expr ==> eq_expr)
                               (apply_fun ψ).
Proof.
  intro e;unfold restr_ax;induction e;intro e';
  destruct e';intuition;simpl in *.
  * intro;repeat rewrite in_map_iff;firstorder. 
  * intro;repeat rewrite in_map_iff;firstorder.
  * unfold is_restr in *;simpl in *.
    assert (τ l =  τ l0).
    destruct l as ((?,?),?);destruct l0 as ((?,?),?);
    unfold τ,eq_letter in *;simpl in *;now inversion H0.
    rewrite H1.
    now apply apply_perm_var_eq_letter.
  * now subst.
  * unfold is_restr in *;simpl in *;repeat simpl_bool.
    split;apply IHe1||apply IHe2;tauto.
  * unfold is_restr in *;simpl in *;repeat simpl_bool.
    split;apply IHe1||apply IHe2;tauto.
  * unfold is_restr in *;simpl in *;repeat simpl_bool.
    apply IHe;tauto.
  * unfold is_restr in *;simpl in *;repeat simpl_bool.
    now firstorder subst.
  * unfold is_restr in *;simpl in *;repeat simpl_bool.
    now firstorder subst.
  * unfold is_restr in *;simpl in *;repeat simpl_bool;split.
    -- now apply IHe.
    -- now rewrite<- equiv_perm_map_ψ.
Qed.

Lemma ψ_surj : forall x, exists y, ψ y = x.
Proof. intro x;exists (φ x);rewrite ψ_φ_id;reflexivity. Qed.
Lemma ψ_inj_implies_φ_surj : (forall x y, ψ x = ψ y -> x = y)
                             -> forall x, X x.
Proof. intros;exists (ψ x);apply H;rewrite ψ_φ_id;reflexivity. Qed.

    
