(** printing $ %\ensuremath{\uparrow}% *)
(** printing £ %\ensuremath{\downarrow}% *)
(** printing α %\ensuremath{\alpha}% *)
(** printing λ %\ensuremath{\lambda}% *)
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Import proofs theories.

Parameter α : Atom.
Parameter λ : Letter.
Axiom τ_λ : τ λ = [α].

Reserved Notation " $ e " (at level 60, no associativity).
Fixpoint single_to_multi exp :=
  match exp with
  | vat a => ‹[a,α]›var λ
  | e + f => ($ e) + $ f
  | e · f => ($ e) · $ f
  | e ^* => ($ e) ^*
  | ν a e => ν a ( $ e )
  | ω a e => ω a ( $ e )
  | ‹ p ›e => ‹ p ›$ e
  | _ => exp
  end
where " $ e " := (single_to_multi e).
Lemma size_single_to_multi e : size ($ e) <= 2 * size e.
Proof. induction e;simpl; lia. Qed.
Lemma mlt_single_to_multi e : multi ($ e).
Proof. induction e;simpl ; tauto. Qed.
(* Lemma ut_mlt_single_to_multi e : *)
(*   untyped e -> multi ($ e). *)
(* Proof. induction e;simpl;tauto. Qed. *)
Lemma ut_ut_single_to_multi e :
  untyped e -> untyped ($ e).
Proof. induction e;simpl;tauto. Qed.
Lemma ut_single_fresh_single_to_multi e :
  untyped e -> single e -> forall a, a # e <-> a # ($ e).
Proof.
  induction e;simpl ;try tauto; intros U S b;try now firstorder.
  rewrite τ_λ;simpl.
  case_eq_at b a;simpl_beq;try tauto.
  case_eq_at b α;simpl_beq;firstorder.
Qed.
Lemma has_type_single_to_multi e A :
  A ⊨ e -> A ⊨ ($ e).
Proof.
  revert A;induction e;simpl;try firstorder.
  intros A ->;rewrite τ_λ;simpl;simpl_beq;reflexivity.
Qed.
Lemma ty_ty_single_to_multi e :
  typed e -> typed ($ e).
Proof. now intros (A&T);exists A;apply has_type_single_to_multi. Qed.
Lemma poa_poa_single_to_multi e :
  str_positive e -> str_positive ($ e).
Proof. induction e;simpl;intuition. Qed.
Lemma po_po_single_to_multi e :
  positive e -> positive ($ e).
Proof. induction e;simpl;intuition
                           (now apply poa_poa_single_to_multi). Qed.

Fixpoint image exp : Prop :=
  match exp with
  | vat _ => False
  | var x => fst (proj1_sig x) = fst (proj1_sig λ)
  | e · f | e + f => image e /\ image f
  | e ^* | ν _ e | ‹ _ › e | ω _ e => image e
  | _ => True
  end.
Lemma sg_image_single_to_multi e :
  single e -> image ($ e).
Proof. induction e;simpl;try tauto. Qed.
(* Lemma image_ut e : image e -> untyped e. *)
(* Proof. induction e;simpl;tauto. Qed. *)
Lemma image_mlt e : image e -> multi e.
Proof. induction e;simpl;tauto. Qed.

Definition get_atom (n : Letter) :=
  match n with
  | exist _ (λ,[a]) _ => a
  | _ => α
  end.

Reserved Notation " £ e " (at level 60, no associativity).
Fixpoint multi_to_single exp :=
  match exp with
  | var x => vat (get_atom x)
  | e + f => (£ e) + £ f
  | e · f => (£ e) · £ f
  | e ^* => (£ e) ^*
  | ν a e => ν a ( £ e )
  | ω a e => ω a ( £ e )
  | ‹ p ›e => ‹ p ›£ e
  | e => e
  end
where " £ e " := (multi_to_single e).

Lemma ut_multi_to_single e :
  (* image e -> *) untyped e -> untyped (£ e).
Proof. induction e;simpl;tauto. Qed.
Lemma img_sg_multi_to_single e :
  image e -> single (£ e).
Proof. induction e;simpl;tauto. Qed.
Lemma img_ut_fresh_multi_to_single e :
  image e -> untyped e -> forall a, a # e <-> a # (£ e).
Proof.
  induction e;simpl ;try tauto; intros h h' b;try now firstorder.
  pose proof τ_λ.
  destruct λ as ((?,?),?).
  destruct l as ((?,?),?).
  unfold τ in *;simpl in *;simpl_is_perm in e0 as x0;
  simpl_is_perm in e as x.
  assert (E:l = x0⋆ ⧓ (x ⧓ [α])).
  now rewrite <-H,e,<-h,<-e0;simpl_perm.
  rewrite E;simpl_in;simpl_perm;simpl.
  now split;intro hyp;[intros eq|intros [eq|I]];auto; apply hyp;
  [left|]; rewrite eq;simpl_perm.
Qed.
Lemma has_type_multi_to_single e A :
  image e -> A ⊨ e -> A ⊨ (£ e).
Proof.
  revert A;induction e;simpl;try firstorder.
  intros A h ->;clear A;pose proof (τ_λ) as h1.
  destruct l as ((?,?),?);destruct λ as ((?,?),?);
  unfold τ in *;simpl in *.
  simpl_is_perm in e as p;simpl_is_perm in e0 as q.
  rewrite <-h in *;clear n0 h;rewrite h1 in *;clear l0 h1.
  rewrite <-e0 in e;clear n e0.
  now assert (l = p⋆ ⧓ (q ⧓  [α])) as ->
      by (now rewrite <- e;simpl_perm).
Qed.
Lemma ty_multi_to_single e :
  image e -> typed e -> typed (£ e).
Proof. now intros h (A&T);exists A;apply has_type_multi_to_single. Qed.

Lemma ka_eq_u_image :
  Proper (ka_eq_u ==> iff) image.
Proof. intros e f E;destruct E;simpl;intuition. Qed.
Lemma ka_eq_t_image :
  Proper (ka_eq_t ==> iff) image.
Proof. intros e f E;destruct E;simpl;intuition. Qed.
Lemma eq_ν_smpl_image :
  Proper (eq_ν_smpl ==> iff) image.
Proof. intros e f E;destruct E;simpl;intuition. Qed.
Lemma eq_nom_untyped_image :
  Proper (eq_nom_untyped ==> iff) image.
Proof. intros e f E;destruct E;simpl;intuition. Qed.
Lemma eq_nom_typed_image :
  Proper (eq_nom_typed ==> iff) image.
Proof. intros e f E;destruct E;simpl;intuition. Qed.
Lemma eq_perm_image :
  Proper (eq_perm ==> iff) image.
Proof. intros e f E;destruct E;simpl;intuition. Qed.
Lemma eq_perm_elim_image :
  Proper (eq_perm_elim ==> iff) image.
Proof. intros e f E;destruct E;simpl;intuition.
       destruct λ as ((?,?),?).
       destruct a as ((?,?),?).
       now simpl in *.
       destruct λ as ((?,?),?).
       destruct a as ((?,?),?).
       now simpl in *.
Qed.
Lemma eq_nom_untyped_p_image :
  Proper (eq_nom_untyped_p ==> iff) image.
Proof. intros e f E;destruct E;simpl;intuition. Qed.
Lemma eq_nom_typed_p_image :
  Proper (eq_nom_typed_p ==> iff) image.
Proof. intros e f E;destruct E;simpl;intuition. Qed.
Lemma NKA_θp_image :
  Proper (NKA_θp ==> iff) image.
Proof.
  intros e f E;repeat destruct E as [E|E] || destruct E;
  simpl;intuition.
  destruct λ as ((?,?),?).
  destruct a as ((?,?),?).
  now simpl in *.
  destruct λ as ((?,?),?).
  destruct a as ((?,?),?).
  now simpl in *.
Qed.
Lemma NKA_θtp_image :
  Proper (NKA_θtp ==> iff) image.
Proof.
  intros e f E;repeat destruct E as [E|E] || destruct E;
  simpl;intuition.
  destruct λ as ((?,?),?).
  destruct a as ((?,?),?).
  now simpl in *.
  destruct λ as ((?,?),?).
  destruct a as ((?,?),?).
  now simpl in *.
Qed.
Lemma eq_NKA_θp_image :
  Proper (eq NKA_θp ==> iff) image.
Proof.
  intros e f E;induction E;try now firstorder.
  now destruct x;destruct y;unfold eq_letter in H;simpl in*;
  intuition subst.
  destruct H;simpl;intuition.
  destruct H;simpl in *;intuition.
  apply NKA_θp_image,H.
Qed.
Lemma eq_NKA_θtp_image :
  Proper (eq NKA_θtp ==> iff) image.
Proof.
  intros e f E;induction E;try now firstorder.
  now destruct x;destruct y;unfold eq_letter in H;simpl in*;
  intuition subst.
  destruct H;simpl;intuition.
  destruct H;simpl in *;intuition.
  apply NKA_θtp_image,H.
Qed.
Lemma NKA_θp_single :
  Proper (NKA_θp ==> iff) single.
Proof.
  intros e f E;repeat destruct E as [E|E] || destruct E;
  simpl;intuition.
Qed.
Lemma NKA_θtp_single :
  Proper (NKA_θtp ==> iff) single.
Proof.
  intros e f E;repeat destruct E as [E|E] || destruct E;
  simpl;intuition.
Qed.
Lemma eq_NKA_θp_single :
  Proper (eq NKA_θp ==> iff) single.
Proof.
  intros e f E;induction E;try now firstorder.
  destruct H;simpl;intuition.
  destruct H;simpl in *;intuition.
  apply NKA_θp_single,H.
Qed.
Lemma eq_NKA_θtp_single :
  Proper (eq NKA_θtp ==> iff) single.
Proof.
  intros e f E;induction E;try now firstorder.
  destruct H;simpl;intuition.
  destruct H;simpl in *;intuition.
  apply NKA_θtp_single,H.
Qed.
Lemma NKA_θp_untyped :
  Proper (NKA_θp ==> iff) untyped.
Proof.
  intros e f E;repeat destruct E as [E|E] || destruct E;
  simpl;intuition.
Qed.
Lemma eq_NKA_θp_untyped :
  Proper (eq NKA_θp ==> iff) untyped.
Proof.
  intros e f E;induction E;try now firstorder.
  destruct H;simpl;intuition.
  destruct H;simpl in *;intuition.
  apply NKA_θp_untyped,H.
Qed.
Hint Resolve eq_NKA_θp_untyped eq_NKA_θp_single eq_NKA_θp_image.
Lemma ut_multi_to_single_single_to_multi e :
  untyped e -> single e -> NKA_θp ⊢ £ $ e ≡ e.
Proof.
  induction e;simpl;intros U S;try now firstorder.
  eql;eql;eql;eqr.
  pose proof τ_λ as T;unfold τ in T; destruct λ as ((?,?),?);
  simpl in *; rewrite T.
  replace a with ([a, α]⋈α) at 2;auto 10.
  simpl;simpl_beq;case_eq_at a α;simpl;simpl_beq;auto.
Qed.
Lemma ty_multi_to_single_single_to_multi e :
  typed e -> single e -> NKA_θtp ⊢ £ $ e ≡ e.
Proof.
  induction e;simpl;intros (A&T) S;try now firstorder.
  reflexivity.
  reflexivity.
  eql;eql;eql;eql;eqr.
  pose proof τ_λ as h;unfold τ in h; destruct λ as ((?,?),?);
  simpl in *; rewrite h.
  replace a with ([a, α]⋈α) at 2;auto 10.
  simpl;simpl_beq;case_eq_at a α;simpl;simpl_beq;auto.
Qed.
Lemma Proper_ut_sg_ax :
  Proper ((ut∧sg∧NKA_θp) ==> eq NKA_θp) single_to_multi.
Proof.
  intros e f (((U1&U2)&(S1&S2))&E);
  repeat destruct E as [E|E]||destruct E;
  simpl in *;try rewrite ut_single_fresh_single_to_multi in *;
  try tauto; auto 10;try now prove_ax.
  * eql;eql;eql;eqr.
    etransitivity;[apply eq_θ;[reflexivity|]|etransitivity];
    try apply eq_ax,eq_θ_var.
    transitivity (var (apply_perm_var (p ++ [a, α]) λ));
      [apply eq_expr_eq;simpl;apply apply_perm_var_app|].
    etransitivity;
      [|symmetry;apply eq_ax,eq_θ_var].
    apply eq_expr_eq;simpl;pose proof τ_λ as T;unfold τ in T.
    unfold eq_letter,apply_perm_var.
    destruct λ as ((?,?),?); simpl in *;rewrite T;simpl;
    repeat f_equal.
    rewrite <-act_comp_app;simpl;simpl_beq;case_eq_at α a;simpl_beq.
    case_eq_at (p⋈a) a;simpl_beq;auto.
    case_eq_at (p⋈a) α;simpl_beq;auto.
Qed.
Lemma Proper_ut_sg_eq :
  Proper ((ut ∧ sg ∧ eq NKA_θp) ==> eq NKA_θp) single_to_multi.
Proof.
  intros e f (((U1&U2)&(S1&S2))&E);revert U1 U2 S1 S2;induction E;
  simpl;intros U1 U2 S1 S2; simpl_hyp;try now firstorder.
  - rewrite IHE1,IHE2;try split;auto;
    eapply eq_NKA_θp_untyped||eapply eq_NKA_θp_single;eauto.
  - destruct H;simpl in *;auto 10.
  - apply (@eq_ka_impl _ ($ e));[apply IHE|];destruct H;simpl in *;
    intuition.
  - apply Proper_ut_sg_ax;repeat split;auto.
Qed.
Lemma Proper_ty_sg_ax :
  Proper ((ty∧sg∧NKA_θtp) ==> eq NKA_θtp) single_to_multi.
Proof.
  intros e f ((((A&T1)&B&T2)&(S1&S2))&E);
  repeat destruct E as [E|E]||destruct E;
  simpl in *;try now prove_ax.
  - now repeat eql;apply eq_ax,kat_t1,has_type_single_to_multi.
  - now repeat eql;apply eq_ax,kat_1t,has_type_single_to_multi.
  - now repeat eql;apply eq_ax,kat_s1,has_type_single_to_multi.
  - now repeat eql;apply eq_ax,kat_s2,has_type_single_to_multi.
  - eql;eql;eql;eql;etransitivity;[eql;eql;eqr;apply eq_ax,eq_θ_θ|].
    eqr;etransitivity;[apply eq_ax,eq_θ_var
                      |symmetry;etransitivity;[apply eq_ax,eq_θ_var|]].
    apply eq_var;pose proof τ_λ as h;destruct λ as ((?,?),?);
    unfold τ, eq_letter in *;simpl in *;rewrite h;f_equal;
    simpl;rewrite <-act_comp_app;simpl;simpl_beq.
    case_eq_at α a;simpl_beq.
    now case_eq_at (p⋈a) a;simpl_beq.
    now case_eq_at (p⋈a) α;simpl_beq.
  - simpl_beq;eql;eqr;apply eq_ax, (@eqt_θν' A0);simpl;intuition.
    now apply has_type_single_to_multi.
  - eqr;apply eq_ax,(@eqt_νω1 A);simpl;repeat simpl_beq;intuition.
    now apply has_type_single_to_multi.
Qed.
Lemma Proper_ty_sg_eq :
  Proper ((ty ∧ sg ∧ eq NKA_θtp) ==> eq NKA_θtp) single_to_multi.
Proof.
  intros e f (((U1&U2)&(S1&S2))&E);revert U1 U2 S1 S2;induction E;
  simpl;intros (A1&T1) (A2&T2) S1 S2; simpl_hyp;try now firstorder.
  - rewrite IHE1,IHE2;try split;auto;eauto;
    try eapply eq_NKA_θtp_single;eauto;
    eexists;eapply (eq_type NKA_θtp_has_type);eauto.
  - now apply eq_bot.
  - now apply eq_id.
  - apply eq_plus;firstorder.
  - apply eq_times;firstorder.
  - destruct H;simpl in *;auto 10.
  - apply (@eq_ka_impl _ ($ e));[apply IHE|];destruct H;simpl in *;
    intuition;eexists;eauto.
  - apply Proper_ty_sg_ax;repeat split;eauto.
Qed.


Definition Img A e f := (image e /\ image f) /\ A e f.
Lemma Proper_ut_img_ax :
  Proper (Img (ut ∧ NKA_θp) ==> eq NKA_θp) multi_to_single.
Proof.
  intros e f (I&((U1&U2)&E));repeat destruct E as [E|E]||destruct E;
  simpl in *;try rewrite img_ut_fresh_multi_to_single in *;
  try tauto; auto 10;try now prove_ax.
  pose proof τ_λ as T;unfold τ in T;
  destruct a as ((?,?),?);destruct λ as ((?,?),?);simpl in *.
  simpl_is_perm in e as x;simpl_is_perm in e0 as x0.
  replace l with (x⋆⧓(x0⧓[α])).
  simpl;eql;eql;eql;eqr;auto.
  now destruct I as (->&_);rewrite <-T,e0,<-e;simpl_perm.
Qed.
Lemma Proper_ut_img_eq :
  Proper (Img (ut ∧ (eq NKA_θp)) ==> eq NKA_θp) multi_to_single.
Proof.
  intros e f (I&(U1&U2)&E);revert I;induction E;simpl;intros I;
  simpl_hyp;try now firstorder.
  - now rewrite IHE1,IHE2;try split;auto;
    apply (eq_untyped NKA_θp_untyped E1)||apply (eq_NKA_θp_image E1).
  - destruct x;destruct y;unfold eq_letter in H;simpl in *;subst;auto.
  - destruct H;simpl in *;auto 10.
  - apply (@eq_ka_impl _ (£ e));[apply IHE|];destruct H;simpl in *;
    intuition.
  - apply Proper_ut_img_ax;repeat split;auto.
Qed.

Lemma Proper_ty_img_ax :
  Proper (Img (ty∧NKA_θtp) ==> eq NKA_θtp) multi_to_single.
Proof.
  intros e f (I&((A1&T1)&A2&T2)&E);
  repeat destruct E as [E|E]||destruct E;
  simpl in *;try now prove_ax.
  - repeat eql;apply eq_ax,kat_t1,has_type_multi_to_single;intuition.
  - repeat eql;apply eq_ax,kat_1t,has_type_multi_to_single;intuition.
  - repeat eql;apply eq_ax,kat_s1,has_type_multi_to_single;intuition.
  - repeat eql;apply eq_ax,kat_s2,has_type_multi_to_single;intuition.
  - pose proof τ_λ as T;unfold τ,eq_letter in *;
    destruct a as ((?,?),?);destruct λ as ((?,?),?);simpl in *.
    simpl_is_perm in e as x;simpl_is_perm in e0 as x0.
    replace l with (x⋆⧓(x0⧓[α])).
    simpl;eql;eql;eql;eql;eqr;auto.
    now destruct I as (->&_);rewrite <-T,e0,<-e;simpl_perm.
  - eql;eqr;eapply eq_ax,(@eqt_θν' A);firstorder.
    now apply has_type_multi_to_single.
  - eqr;eapply eq_ax,(@eqt_νω1 A);firstorder.
    now apply has_type_multi_to_single;simpl;repeat simpl_beq.
Qed.

Lemma Proper_ty_img_eq :
  Proper (Img (ty ∧ (eq NKA_θtp)) ==> eq NKA_θtp) multi_to_single.
Proof.
  unfold ty,rel,typed;intros e f (I&U&E);revert I U;induction E;simpl;
  intros I ((A1&T1)&A2&T2);simpl_hyp;try now firstorder.
  - rewrite IHE1,IHE2;try split;auto; try apply (eq_NKA_θtp_image E1);
    auto;eexists;eauto; eapply (eq_type NKA_θtp_has_type _ E1);eauto.
  - now apply eq_bot.
  - now apply eq_id.
  - destruct x;destruct y;unfold eq_letter in H;simpl in *;subst;auto.
  - apply eq_plus; firstorder.
  - apply eq_times; firstorder.
  - destruct H;simpl in *;auto 10.
  - apply (@eq_ka_impl _ (£ e));[apply IHE|];destruct H;simpl in *;
    intuition;eauto.
  - apply Proper_ty_img_ax;repeat split;eauto.
Qed.
