Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Require Import proofs rm0 removetype perm.


(** * Equationnal theories. *)

Definition TyEq :=  (predicate expr *relation expr)%type.

Definition NKA_θtp :=
  (ka_eq_t∨eq_perm∨eq_perm_ω∨eq_perm_elim
          ∨eq_ν_smpl∨eq_ω_smpl∨eq_nom_typed_p∨eq_nom_typed).
Definition NKAmpt : TyEq := (multi⊗str_positive⊗typed,NKA_θtp).
Definition NKAspt : TyEq  := (single⊗str_positive⊗typed,NKA_θtp).

Definition NKA_θt :=NKA_θtp∨zero_eq_t.
Definition NKAmt : TyEq := (multi⊗typed,NKA_θt).
Definition NKAst : TyEq := (single⊗typed,NKA_θt).

Definition NKA_θp :=
  (ka_eq_u∨eq_perm∨eq_perm_elim
          ∨eq_ν_smpl∨eq_nom_untyped_p∨eq_nom_untyped).
Definition NKAmpu : TyEq := (multi⊗str_positive⊗untyped,NKA_θp).
Definition NKAspu : TyEq := (single⊗str_positive⊗untyped,NKA_θp).

Definition NKA_θ := NKA_θp∨zero_eq_u.
Definition NKAmu : TyEq := (multi⊗untyped,NKA_θ).
Definition NKAsu : TyEq := (single⊗untyped,NKA_θ).

Definition NKA_tp :=
  (ka_eq_t∨eq_ν_smpl∨eq_ω_smpl∨eq_nom_typed∨eq_nom_typed_np).
Definition NKAnspt : TyEq := (no_perm⊗single⊗str_positive⊗typed,NKA_tp).
Definition NKAnmpt : TyEq := (no_perm⊗multi⊗str_positive⊗typed,NKA_tp).

Definition NKA_t := NKA_tp∨zero_eq_t.
Definition NKAnst : TyEq := (no_perm⊗single⊗typed,NKA_t).
Definition NKAnmt : TyEq := (no_perm⊗multi⊗typed,NKA_t).

Definition NKA_p :=
  (ka_eq_u∨eq_ν_smpl∨eq_nom_untyped∨eq_nom_untyped_np).
Definition NKAnspu : TyEq :=
  (no_perm⊗single⊗str_positive⊗untyped,NKA_p).
Definition NKAnmpu : TyEq := (no_perm⊗multi⊗str_positive⊗untyped,NKA_p).

Definition NKA := NKA_p∨zero_eq_u.
Definition NKAnsu : TyEq := (no_perm⊗single⊗untyped,NKA).
Definition NKAnmu : TyEq := (no_perm⊗multi⊗untyped,NKA).

(** The following sets of axioms will only be used as *)
(** intermediary in proofs.*)
Definition NKA_tp' :=
  (((ka_eq_t ∨ eq_ν_smpl) ∨ eq_ω_smpl) ∨ eq_nom_typed)
    ∨ (np ∧ eq_nom_typed_np).
Definition NKA_p' :=
  ((ka_eq_u ∨ eq_ν_smpl) ∨ eq_nom_untyped)
    ∨ (np ∧ eq_nom_untyped_np).
Definition NKA' :=
  ut ∧ NKA_p.


(** * Simple Properties of these theories. *)
(** ** Preservation of [single]. *)
Lemma NKA_p_single : Proper (NKA_p ==> iff) single.
Proof.
  intros e f [[[h|h]|h]|h].
  now apply ka_eq_u_single.
  now apply eq_ν_smpl_single.
  now apply eq_nom_untyped_single.
  now apply eq_nom_untyped_np_single.
Qed.
Lemma NKA_tp_single : Proper (NKA_tp ==> iff) single.
Proof.
  intros e f [[[[h|h]|h]|h]|h].
  now apply ka_eq_t_single.
  now apply eq_ν_smpl_single.
  now apply eq_ω_smpl_single.
  now apply eq_nom_typed_single.
  now apply eq_nom_typed_np_single.
Qed.
Lemma NKA_θp_single : Proper (NKA_θp ==> iff) single.
Proof.
  intros e f [[[[[h|h]|h]|h]|h]|h].
  now apply ka_eq_u_single.
  now apply eq_perm_single.
  now apply eq_perm_elim_single.
  now apply eq_ν_smpl_single.
  now apply eq_nom_untyped_p_single.
  now apply eq_nom_untyped_single.
Qed.
Lemma NKA_θtp_single : Proper (NKA_θtp ==> iff) single.
Proof.
  intros e f [[[[[[[h|h]|h]|h]|h]|h]|h]|h].
  now apply ka_eq_t_single.
  now apply eq_perm_single.
  now apply eq_perm_ω_single.
  now apply eq_perm_elim_single.
  now apply eq_ν_smpl_single.
  now apply eq_ω_smpl_single.
  now apply eq_nom_typed_p_single.
  now apply eq_nom_typed_single.
Qed.
(** ** Preservation of [multi]. *)
Lemma NKA_p_multi : Proper (NKA_p ==> iff) multi.
Proof.
  intros e f [[[h|h]|h]|h].
  now apply ka_eq_u_multi.
  now apply eq_ν_smpl_multi.
  now apply eq_nom_untyped_multi.
  now apply eq_nom_untyped_np_multi.
Qed.
Lemma NKA_tp_multi : Proper (NKA_tp ==> iff) multi.
Proof.
  intros e f [[[[h|h]|h]|h]|h].
  now apply ka_eq_t_multi.
  now apply eq_ν_smpl_multi.
  now apply eq_ω_smpl_multi.
  now apply eq_nom_typed_multi.
  now apply eq_nom_typed_np_multi.
Qed.
Lemma NKA_θp_multi : Proper (NKA_θp ==> iff) multi.
Proof.
  intros e f [[[[[h|h]|h]|h]|h]|h].
  now apply ka_eq_u_multi.
  now apply eq_perm_multi.
  now apply eq_perm_elim_multi.
  now apply eq_ν_smpl_multi.
  now apply eq_nom_untyped_p_multi.
  now apply eq_nom_untyped_multi.
Qed.
Lemma NKA_θtp_multi : Proper (NKA_θtp ==> iff) multi.
Proof.
  intros e f [[[[[[[h|h]|h]|h]|h]|h]|h]|h].
  now apply ka_eq_t_multi.
  now apply eq_perm_multi.
  now apply eq_perm_ω_multi.
  now apply eq_perm_elim_multi.
  now apply eq_ν_smpl_multi.
  now apply eq_ω_smpl_multi.
  now apply eq_nom_typed_p_multi.
  now apply eq_nom_typed_multi.
Qed.
(** ** Preservation of [str_positive]. *)
Lemma NKA_p_str_positive : Proper (NKA_p ==> iff) str_positive.
Proof.
  intros e f [[[h|h]|h]|h];destruct h;simpl;try now firstorder.
  apply str_positive_apply_perm.
Qed.
Lemma NKA_tp_str_positive : Proper (NKA_tp ==> iff) str_positive.
Proof.
  intros e f [[[[h|h]|h]|h]|h];destruct h;simpl;try now firstorder.
  apply str_positive_apply_perm. 
Qed.
Lemma NKA_θp_str_positive : Proper (NKA_θp ==> iff) str_positive.
Proof.
  intros e f [[[[[h|h]|h]|h]|h]|h];destruct h;simpl;firstorder.
Qed.
Lemma NKA_θtp_str_positive : Proper (NKA_θtp ==> iff) str_positive.
Proof.
  intros e f [[[[[[[h|h]|h]|h]|h]|h]|h]|h];destruct h;simpl;firstorder.
Qed.
(** ** Type preservation. *)
Lemma NKA_θtp_has_type A : Proper (NKA_θtp ==> iff) (has_type A).
Proof. unfold NKA_θtp;split_proper;auto with propertype. Qed.
Lemma NKA_θt_has_type A : Proper (NKA_θt ==> iff) (has_type A).
Proof. unfold NKA_θt,NKA_θtp;split_proper; auto with propertype. Qed.
Lemma NKA_tp_has_type A : Proper (NKA_tp ==> iff) (has_type A).
Proof. unfold NKA_tp;split_proper;auto with propertype. Qed.
Lemma NKA_t_has_type A : Proper (NKA_t ==> iff) (has_type A).
Proof. unfold NKA_t,NKA_tp;split_proper; auto with propertype. Qed.
Lemma NKA_θp_untyped : Proper (NKA_θp ==> iff) untyped.
Proof. unfold NKA_θp;split_proper;auto with propertype. Qed.
Lemma NKA_θ_untyped : Proper (NKA_θ ==> iff) untyped.
Proof. unfold NKA_θ,NKA_θp;split_proper;auto with propertype. Qed.
Lemma NKA_p_untyped : Proper (NKA_p ==> iff) untyped.
Proof. unfold NKA_p;split_proper;auto with propertype. Qed.
Lemma NKA_untyped : Proper (NKA ==> iff) untyped.
Proof. unfold NKA,NKA_p;split_proper;auto with propertype. Qed.
Lemma NKA_tp'_has_type A : Proper (NKA_tp' ==> iff) (has_type A).
Proof.
  unfold NKA_tp';split_proper;
  repeat apply (proper_stronger_l (@and_stronger_l _ _ _));
  auto with propertype.
Qed.
Lemma NKA_p'_untyped : Proper (NKA_p' ==> iff) untyped.
Proof.
  unfold NKA_p';split_proper;
  repeat apply (proper_stronger_l (@and_stronger_l _ _ _));
  auto with propertype.
Qed.
Lemma NKA'_untyped : Proper (NKA' ==> iff) untyped.
Proof. unfold NKA',ut,rel,and_rel,Proper,respectful;firstorder. Qed.

(** ** NKA' is not different from NKA_p. *)
Lemma switch_NKA'_NKA_p :
  forall e f, ut e f ->  NKA_p ⊢ e ≡ f -> NKA' ⊢ e ≡ f.
Proof.
  intros e f (u&u')  h;apply remain_Ut;auto.
  apply NKA_p_untyped.
Qed.
Lemma switch_NKA_p_NKA' :
  forall e f, NKA' ⊢ e ≡ f -> NKA_p ⊢ e ≡ f.
Proof.
  intros e f;apply stronger_ax_proof.
  unfold NKA';firstorder.
Qed.
(** ** Preservation of [eq0]. *)
Lemma NKA_θtp_eq0 : Proper (NKA_θtp ==> Logic.eq) eq0.
Proof. unfold NKA_θtp;split_proper;auto with propereq0. Qed.
Lemma NKA_θt_eq0 : Proper (NKA_θt ==> Logic.eq) eq0.
Proof. unfold NKA_θt,NKA_θtp;split_proper; auto with propereq0. Qed.
Lemma NKA_tp_eq0 : Proper (NKA_tp ==> Logic.eq) eq0.
Proof. unfold NKA_tp;split_proper;auto with propereq0. Qed.
Lemma NKA_t_eq0 : Proper (NKA_t ==> Logic.eq) eq0.
Proof. unfold NKA_t,NKA_tp;split_proper; auto with propereq0. Qed.
Lemma NKA_θp_eq0 : Proper (NKA_θp ==> Logic.eq) eq0.
Proof. unfold NKA_θp;split_proper;auto with propereq0. Qed.
Lemma NKA_θ_eq0 : Proper (NKA_θ ==> Logic.eq) eq0.
Proof. unfold NKA_θ,NKA_θp;split_proper;auto with propereq0. Qed.
Lemma NKA_p_eq0 : Proper (NKA_p ==> Logic.eq) eq0.
Proof. unfold NKA_p;split_proper;auto with propereq0. Qed.
Lemma NKA_eq0 : Proper (NKA ==> Logic.eq) eq0.
Proof. unfold NKA,NKA_p;split_proper;auto with propereq0. Qed.
(** ** Preservation of [rm0]. *)
Ltac split_proper :=
  repeat (try apply (proper_stronger_l (@and_or_distr _ _ _ _));
           rewrite proper_or_split;split).
Lemma NKA_θtp_rm0 : Proper (ty' ∧ NKA_θtp ==> eq NKA_θtp) rm0.
Proof.
  unfold NKA_θtp;split_proper;intros e f (T&h).
  - repeat eql; apply rm0_t_ka_eq;split;auto.
  - pose proof (@rm0_eq_perm_elim e f) as h'.
    unfold or_rel in *;firstorder; auto 10;rewrite H0;auto.
  - apply rm0_eq_perm_ω in h.
    unfold or_rel in *;firstorder; auto 10;rewrite H1;auto.
  - pose proof (@rm0_eq_perm_elim e f) as h'.
    unfold or_rel in *;firstorder; auto 10;rewrite H;auto.
  - pose proof (@rm0_eq_ν_smpl_typed e f) as h';
    eql;eql;eql;eqr;unfold or_rel in *;firstorder.
  - pose proof (@rm0_eq_ω_smpl_typed e f) as h';
    eql;eql;eqr;unfold or_rel in *;firstorder.
  - apply rm0_eq_nom_typed_p in h;eql;eqr;auto 10.
  - pose proof (@rm0_eq_nom_typed e f) as h';
    eqr;unfold or_rel in *;firstorder.
Qed.
Lemma NKA_tp_rm0 : Proper (ty' ∧ NKA_tp ==> eq NKA_tp) rm0.
Proof.
  unfold NKA_tp;split_proper;intros e f (T&h).
  - repeat eql; apply rm0_t_ka_eq;split;auto.
  - pose proof (@rm0_eq_ν_smpl_typed e f) as h';
    eql;eql;eql;eqr;unfold or_rel in *;firstorder.
  - pose proof (@rm0_eq_ω_smpl_typed e f) as h';
    eql;eql;eqr;unfold or_rel in *;firstorder.
  - pose proof (@rm0_eq_nom_typed e f) as h';
    eql;eqr;unfold or_rel in *;firstorder.
  - pose proof (@rm0_eq_nom_typed_np e f) as h';
    eqr;unfold or_rel in *;firstorder.
Qed.
Lemma NKA_θp_rm0 : Proper (ut ∧ NKA_θp ==> eq NKA_θp) rm0.
Proof.
  unfold NKA_θp;split_proper;intros e f (T&h).
  - repeat eql; apply rm0_u_ka_eq;split;auto.
  - pose proof (@rm0_eq_perm_elim e f) as h'.
    unfold or_rel in *;firstorder; auto 10 ;rewrite H2;auto.
  - pose proof (@rm0_eq_perm_elim e f) as h'.
    unfold or_rel in *;firstorder; auto 10;rewrite H1;auto.
  - pose proof (@rm0_eq_ν_smpl_untyped e f) as h';
    eql;eql;eqr;unfold or_rel in *;firstorder.
  - eql;eqr;apply rm0_eq_nom_untyped_p;split;auto. 
  - pose proof (@rm0_eq_nom_untyped e f) as h';
    eqr;unfold or_rel in *;firstorder.
Qed.
Lemma NKA_p_rm0 : Proper (ut ∧ NKA_p ==> eq NKA_p) rm0.
Proof.
  unfold NKA_p;split_proper;intros e f (T&h).
  - repeat eql; apply rm0_u_ka_eq;split;auto.
  - pose proof (@rm0_eq_ν_smpl_untyped e f) as h';
    eql;eql;eqr;unfold or_rel in *;firstorder.
  - pose proof (@rm0_eq_nom_untyped e f) as h';
    eql;eqr;unfold or_rel in *;firstorder.
  - pose proof (@rm0_eq_nom_untyped_np e f) as h';
    eqr;unfold or_rel in *;firstorder.
Qed.

(** ** [support] preservation. *)
Lemma NKA_θp_support : Proper (NKA_θp ==> equiv_lst) support.
Proof. unfold NKA_θp;split_proper;auto with propersupport. Qed.
Lemma NKA_p_support : Proper (NKA_p ==> equiv_lst) support.
Proof. unfold NKA_p;split_proper;auto with propersupport. Qed.
Lemma NKA_p'_support : Proper (NKA_p' ==> equiv_lst) support.
Proof.
  apply proper_stronger_l with (A:=NKA_p);[firstorder|];
  apply NKA_p_support.
Qed.
Lemma NKA'_support : Proper (NKA' ==> equiv_lst) support.
Proof.
  apply proper_stronger_l with (A:=NKA_p);[firstorder|];
  apply NKA_p_support.
Qed.

(** ** Preservation of [erase]. *)
Lemma NKAθ_erase : Proper (ty' ∧ NKA_θtp ==> eq NKA_θp) erase.
Proof.
  unfold NKA_θp;split_proper;intros e f (T&h).
  - apply ka_eq_tu_erase in h;repeat eql;auto.
  - apply eq_perm_erase in h;eql;eql;eql;eql;eqr;auto.
  - apply eq_perm_ω_erase in h as ->;auto.
  - apply eq_perm_elim_erase in h;eql;eql;eql;eqr;auto.
  - apply eq_ν_smpl_erase in h;eql;eql;eqr;auto.
  - apply eq_ω_smpl_erase in h as ->;auto.
  - eapply stronger_ax_proof;[|apply eq_nom_p_erase];firstorder.
  - eapply stronger_ax_proof;[|apply eq_nom_erase];firstorder.
Qed.
Lemma NKA_erase' : Proper (ty' ∧ NKA_tp' ==> eq NKA_p') erase.
Proof.
  unfold NKA_p';split_proper;intros e f (T&h).
  - apply ka_eq_tu_erase in h;repeat eql;auto.
  - apply eq_ν_smpl_erase in h;eql;eql;eqr;auto.
  - apply eq_ω_smpl_erase in h as ->;auto.
  - eapply stronger_ax_proof;[|apply eq_nom_erase];firstorder.
  - apply eq_nom_np_erase in h;eqr;auto.
Qed.
(** ** Preservation of [typed_vers]. *)
Lemma NKAθ_typed_vers : Proper (ut ∧ NKA_θp ==> eq NKA_θtp) typed_vers.
Proof.
  unfold NKA_θtp;split_proper;intros e f h.
  - eapply stronger_ax_proof;[|apply ka_eq_tu_typed_vers];firstorder.
  - eapply stronger_ax_proof;[|apply eq_perm_typed_vers];firstorder.
  - destruct h as (u&h);apply eq_perm_elim_typed_vers in h;auto 10.
  - eapply stronger_ax_proof;[|apply eq_ν_smpl_typed_vers];firstorder.
  - eapply stronger_ax_proof;[|apply eq_nom_typed_p_vers];firstorder.
  - eapply stronger_ax_proof;[|apply eq_nom_typed_vers];firstorder.
Qed.
Lemma NKA_typed_vers' : Proper (ut ∧ NKA_p' ==> eq NKA_tp') typed_vers.
Proof.
  unfold NKA_tp';split_proper;intros e f h.
  - eapply stronger_ax_proof;[|apply ka_eq_tu_typed_vers];firstorder.
  - eapply stronger_ax_proof;[|apply eq_ν_smpl_typed_vers];firstorder.
  - eapply stronger_ax_proof;[|apply eq_nom_typed_vers];firstorder.
  - eapply stronger_ax_proof;[|apply eq_nom_typed_np_vers];firstorder.
Qed.

(** ** Preservation of [str_positive]. *)
Lemma NKA'_pos : forall e f, NKA' ⊢ e ≡ f ->
                             (str_positive e <-> str_positive f).
Proof.
  intros e f E;induction E;simpl in *;try tauto.
  destruct H;simpl;tauto.
  rewrite<- IHE at 2;destruct H;simpl;tauto.
  destruct H as (U&H);
    repeat destruct H as [H|H];try destruct H;simpl;try tauto.
  rewrite <-str_positive_apply_perm;tauto.
Qed.
Lemma NKA_pos : forall e f, NKA_p ⊢ e ≡ f ->
                             (str_positive e <-> str_positive f).
Proof.
  intros e f E;induction E;simpl in *;try tauto.
  destruct H;simpl;tauto.
  rewrite<- IHE at 2;destruct H;simpl;tauto.
  repeat destruct H as [H|H];try destruct H;simpl;try tauto.
  rewrite <-str_positive_apply_perm;tauto.
Qed.
(** ** Preservation of [no_perm]. *)
Lemma NKA_p_Np : Proper (NKA_p ==> iff) no_perm.
Proof.
  intros e f [[[h|h]|h]|h].
  now apply ka_eq_u_no_perm.
  now apply eq_ν_smpl_no_perm.
  now apply eq_nom_untyped_no_perm.
  apply eq_nom_untyped_np_no_perm in h;tauto.
Qed.
Lemma NKA_tp_Np : Proper (NKA_tp ==> iff) no_perm.
Proof.
  intros e f [[[[h|h]|h]|h]|h].
  now apply ka_eq_t_no_perm.
  now apply eq_ν_smpl_no_perm.
  now apply eq_ω_smpl_no_perm.
  now apply eq_nom_typed_no_perm.
  apply eq_nom_typed_np_no_perm in h;tauto.
Qed.
(* (** ** Preservation of [single⊗no_perm]. *) *)
(* Lemma NKA_tp_SgNp : Proper (NKA_tp ==> iff) (single ⊗ no_perm). *)
(* Proof. *)
(*   unfold and_pred;intros e f [[[[h|h]|h]|h]|h]. *)
(*   now rewrite (ka_eq_t_single h),(ka_eq_t_no_perm h). *)
(*   now rewrite (eq_ν_smpl_single h),(eq_ν_smpl_no_perm h). *)
(*   now rewrite (eq_ω_smpl_single h),(eq_ω_smpl_no_perm h). *)
(*   now rewrite (eq_nom_typed_single h),(eq_nom_typed_no_perm h). *)
(*   now apply eq_nom_typed_np_no_perm. *)
(* Qed. *)
(* Lemma NKA_p_SgNp : Proper (NKA_p ==> iff) (single⊗no_perm). *)
(* Proof. *)
(*   unfold and_pred;intros e f [[[h|h]|h]|h]. *)
(*   now rewrite (ka_eq_u_single h),(ka_eq_u_no_perm h). *)
(*   now rewrite (eq_ν_smpl_single h),(eq_ν_smpl_no_perm h). *)
(*   now rewrite (eq_nom_untyped_single h),(eq_nom_untyped_no_perm h). *)
(*   now apply eq_nom_untyped_np_no_perm. *)
(* Qed. *)
(* Lemma eqNKA_tp_SgNp : Proper (eq NKA_tp ==> iff) (single⊗no_perm). *)
(* Proof. *)
(*   unfold and_pred;intros e f π;induction π;simpl in *;try tauto. *)
(*   now rewrite (ka_eq_single H),(ka_eq_no_perm H). *)
(*   destruct H;simpl in *;tauto. *)
(*   now apply NKA_tp_SgNp. *)
(* Qed. *)
(* Lemma eqNKA_p_SgNp : Proper (eq NKA_p ==> iff) (single⊗no_perm). *)
(* Proof. *)
(*   unfold and_pred;intros e f π;induction π;simpl in *;try tauto. *)
(*   now rewrite (ka_eq_single H),(ka_eq_no_perm H). *)
(*   destruct H;simpl in *;tauto. *)
(*   now apply NKA_p_SgNp. *)
(* Qed. *)
(** ** Zero is different than One. *)
Lemma zero_diff_un_NKA_θp : ~ (NKA_θp ⊢ 0 ≡ 1).
Proof.
  intros h;apply eq_eq0 in h;[discriminate|].
  unfold NKA_θp;split_proper; auto with propereq0.
Qed.
Lemma zero_diff_un_NKA_θtp : ~ (NKA_θtp ⊢ ⊥ [] ≡ id []).
Proof.
  intros h;apply eq_eq0 in h;[discriminate|].
  unfold NKA_θtp;split_proper;auto with propereq0.
Qed.
Lemma zero_diff_un_NKA_p : ~ (NKA_p ⊢ 0 ≡ 1).
Proof.
  intros h;apply eq_eq0 in h;[discriminate|].
  unfold NKA_p;split_proper; auto with propereq0.
Qed.
Lemma zero_diff_un_NKA_tp : ~ (NKA_tp ⊢ ⊥ [] ≡ id []).
Proof.
  intros h;apply eq_eq0 in h;[discriminate|].
  unfold NKA_tp;split_proper;auto with propereq0.
Qed.
Lemma at_diff_un_NKA_θp a : ~ (NKA_θp ⊢ vat a ≡ 1).
Proof.
  intros h;apply eq_support in h;simpl in h;
  [unfold equiv_lst in h;simpl in h;rewrite <-h;eauto|].
  unfold NKA_θp;split_proper; auto with propersupport.
Qed.
Lemma le_diff_un_NKA_θp x : τ x <> [] -> ~ (NKA_θp ⊢ var x ≡ 1).
Proof.
  intros hyp h;apply eq_support in h;simpl in h;
  [unfold equiv_lst in h;simpl in h;destruct (τ x);simpl;
   firstorder;rewrite <- h;left;eauto|].
  unfold NKA_θp;split_proper; auto with propersupport.
Qed.
Lemma id_a_diff_un_NKA_θtp a : ~ (NKA_θtp ⊢ id [a] ≡ id []).
Proof.
  intros h;apply eq_type with (A := []) in h;simpl in h.
  - unfold equiv_lst in h;simpl in h;intuition.
    rewrite H1;eauto.
  - unfold NKA_θtp;intro;split_proper;auto with propertype.
Qed.
Lemma at_diff_un_NKA_p a : ~ (NKA_p ⊢ vat a ≡ 1).
Proof.
  intros h;apply eq_support in h;simpl in h;
  [unfold equiv_lst in h;simpl in h;rewrite <-h;eauto|].
  unfold NKA_p;split_proper; auto with propersupport.
Qed.
Lemma le_diff_un_NKA_p x : τ x <> [] -> ~ (NKA_p ⊢ var x ≡ 1).
Proof.
  intros hyp h;apply eq_support in h;simpl in h;
  [unfold equiv_lst in h;simpl in h;destruct (τ x);simpl;
   firstorder;rewrite <- h;left;eauto|].
  unfold NKA_p;split_proper; auto with propersupport.
Qed.
Lemma id_a_diff_un_NKA_tp a : ~ (NKA_tp ⊢ id [a] ≡ id []).
Proof.
  intros h;apply eq_type with (A := []) in h;simpl in h.
  - unfold equiv_lst in h;simpl in h;intuition.
    rewrite H1;eauto.
  - unfold NKA_tp;intro;split_proper;auto with propertype.
Qed.

(** ** Transfer lemmas from one system to the other. *)
(** [NKA_tp'] is equivalent to [NKA_tp] for typed expressions 
 without permutation over single variables. *)
Lemma NKA_tp'_NKA_tp_eq :
  forall e f,(np ∧ ty) e f ->
             (NKA_tp'⊢ e≡f <-> NKA_tp⊢ e≡f).
Proof.
  intros e f ((N1&N2)&(A&T1)&B&T2);split;intro π.
  eapply stronger_ax_proof;[|apply π].
  unfold NKA_tp,NKA_tp',np,rel,and_rel,or_rel,stronger;
    intuition.
  revert A B T1 T2 N1 N2;induction π;intros X Y T1 T2 N1 N2;
  auto.
  - symmetry;firstorder.
  - assert (T3: X⊨f).
    pose proof (eq_type NKA_tp_has_type X π1) as E;rewrite <-E;apply T1.
    now rewrite IHπ1,IHπ2;eauto;rewrite (eq_no_perm NKA_tp_Np π2).
  - rewrite (IHπ1 X),(IHπ2 X);simpl in *;auto;firstorder;
    simpl_hyp;eauto.
  - rewrite (IHπ1 X),(IHπ2 X);simpl in *;auto;firstorder;
    simpl_hyp;eauto.
  - rewrite (IHπ X);simpl in *;auto;firstorder;
    simpl_hyp;eauto.
  - rewrite (IHπ (a::X));simpl in *;auto;firstorder;
    simpl_hyp;eauto.
  - rewrite (IHπ (p⋆⧓X));simpl in *;auto;firstorder;
    simpl_hyp;eauto.
  - rewrite (IHπ (rm a X));simpl in *;auto;firstorder;
    simpl_hyp;eauto.
  - eapply eq_ka_impl;[apply (IHπ X Y)|apply H];
    destruct H;simpl in *;firstorder.
  - repeat destruct H as [H|H]. 
    * repeat eql;auto.
    * eql;eql;eql;eqr;auto.
    * eql;eql;eql;eqr;auto.
    * eql;eql;eqr;auto.
    * eql;eqr;auto.
    * eqr;auto 10.
Qed.

(** [NKA_p'] is equivalent to [NKA_p] for untyped expressions 
 without permutation over single variables. *)
Lemma NKA_p'_NKA_p_eq :
  forall e f,(np∧ut) e f ->
             (NKA_p'⊢ e≡f <-> NKA_p⊢ e≡f).
Proof.
  intros e f ((N1&N2)&U1&U2);split;intro π.
  eapply stronger_ax_proof;[|apply π].
  unfold NKA_p,NKA_p',np,rel,and_rel,or_rel,stronger;intuition.
  induction π;auto.
  - assert (U3:untyped f).
    pose proof (eq_untyped NKA_p_untyped π1) as E;rewrite <-E;apply U1.
    now rewrite IHπ1,IHπ2;eauto;rewrite (eq_no_perm NKA_p_Np π2).
  - rewrite IHπ1,IHπ2;simpl in *;auto;firstorder.
  - rewrite IHπ1,IHπ2;simpl in *;auto;firstorder.
  - rewrite IHπ;simpl in *;auto;firstorder.
  - rewrite IHπ;simpl in *;auto;firstorder.
  - eapply eq_ka_impl;[apply IHπ|apply H];destruct H;simpl in *;
    firstorder.
  - repeat destruct H as [H|H].
    * repeat eql;auto.
    * eql;eql;eqr;auto.
    * eql;eql;eqr;auto.
    * eql;eqr;auto.
    * eqr;auto 10.
Qed.
(** For positive untyped expressions, [eq NKA'] entails [eq NKA_θp]. *)
Lemma NKA'_NKA_θp e f :
  NKA' ⊢ e ≡ f -> positive e -> positive f ->
  untyped e -> untyped f ->
  NKA_θp ⊢ e ≡ f.
Proof.
  intros π p1' p2' u1 u2.
  assert (e = 0 \/ str_positive e)
    as p1 by (destruct e;simpl;auto);clear p1'.
  assert (f = 0 \/ str_positive f)
    as p2 by (destruct f;simpl;auto);clear p2'.
  destruct p1 as [->|p1];destruct p2 as [->|p2].
  reflexivity.
  rewrite<- (NKA'_pos π) in p2;simpl in p2;tauto.
  rewrite (NKA'_pos π) in p1;simpl in p1;tauto.
  induction π;simpl in *;auto.
  - now rewrite IHπ1,IHπ2;simpl;auto;try rewrite NKA'_pos;
    eauto;apply (eq_untyped NKA'_untyped π2).
  - rewrite IHπ1,IHπ2;simpl;try tauto;auto.
  - rewrite IHπ1,IHπ2;simpl;try tauto;auto.
  - rewrite IHπ;simpl;try tauto;auto.
  - apply IHπ in u2;try tauto.
    apply (eq_ka_impl u2);auto.
    cut (untyped (e + g));try (simpl;tauto).
    now apply (eq_untyped NKA'_untyped π).
    cut (str_positive (e + g));try (simpl;tauto).
    rewrite NKA'_pos;eauto.
  - destruct H as ((U1&U2)&E);
    repeat destruct E as [E|E];simpl;auto.
    * repeat eql;auto.
    * eql;eql;eqr;auto.
    * eql;eql;eqr;auto.
    * eqr;auto.
    * assert (E':(ut ∧ eq_nom_untyped_np) e f)
        by (now repeat split).
       apply eq_nom_untyped_np_rm_perm_pos in E';auto.
       eapply stronger_ax_proof;[|apply E'].
       unfold NKA_θp,or_rel,sg,rel,and_rel,stronger;intuition.
Qed.

Lemma NKA_tp_NKA_θtp e f:
  ty e f /\ sp e f /\ np e f ->
  NKA_tp ⊢ e ≡ f -> NKA_θtp ⊢ e ≡ f. 
Proof.
  pose proof (eq_type NKA_tp_has_type) as ET1.
  pose proof (eq_type NKA_θtp_has_type) as ET2.
  pose proof (eq_str_positive NKA_tp_str_positive) as ES1.
  pose proof (eq_str_positive NKA_θtp_str_positive) as ES2.
  pose proof (eq_no_perm NKA_tp_Np) as EN.
  intros (((A&T1)&(B&T2))&(S1&S2)&(N1&N2)).
  intro E;revert A B T1 T2;induction E.
  - firstorder.
  - intros A B T1 T2;etransitivity;eapply IHE1 ||eapply IHE2;
     try apply S1 || apply S2 || apply T1 || apply T2 || apply N1
     || apply N2 || apply (ES1 _ _ E1)|| apply (ET1 B _ _ E2)
     || apply (EN _ _ E2);auto.
  - firstorder.
  - firstorder.
  - firstorder.
  - firstorder. exact 1.
  - firstorder. exact 1.
  - firstorder. exact 1.  
  - simpl in *;intros A B (T1&T2) (T3 & T4);rewrite IHE1, IHE2;
    reflexivity||tauto||eauto.
  - simpl in *;intros A B (T1&T2) (T3 & T4);rewrite IHE1, IHE2;
    reflexivity||tauto||eauto.
  - simpl in *;intros A B T1 T2;rewrite IHE;
    reflexivity||tauto||eauto.
  - simpl in *;intros A B (T1&T2) (T3 & T4);rewrite IHE;
    reflexivity||tauto||eauto.
  - simpl in *;tauto. 
  - simpl in *;intros A B (T1&T2) (T3 & T4);rewrite IHE;
    reflexivity||tauto||eauto.
  - now intros;apply eq_ka.
  - intros A B (T1&T2) T3;eapply eq_ka_impl;[eapply IHE|assumption];
    rewrite<- (ET1 A _ _ E) in T2;rewrite<- (ES1 _ _ E) in S2;
    rewrite<- (EN _ _ E) in N2;simpl in *;try tauto;eauto.
  - repeat destruct H as [H|H];try destruct H;intros B C T1 T2;
    try now prove_ax.
    etransitivity;[eql;eqr;apply eq_ax,(@eqt_θν' A _ b);auto|].
    eapply eq_ν,stronger_ax_proof;[|apply eq_apply_perm_pos].
    unfold NKA_θtp,or_rel,stronger;intuition.
    now simpl in *;  apply str_positive_apply_perm in S2.
Qed.

(** For every permutation [p], the function [rm_perm ∘ apply_perm p] is 
a morphism between [NKA_θp] and [eq (NKA_p)] *)
Lemma NKA_θp_NKA_p_rm_perm_apply_perm :
  forall p : perm,
    Proper (NKA_θp ==> eq NKA_p)
           (rm_perm ∘ apply_perm p).
Proof.
  unfold Basics.compose;
  intros p e f [[[[[E|E]|E]|E]|E]|E].
  * apply eq_ax.
     destruct E;simpl;repeat left;auto 10.
  * now apply (stronger_ax_proof (FFalse_spec _)),eq_perm_apply_perm.
  * now apply (stronger_ax_proof (FFalse_spec _)),
     eq_perm_elim_apply_perm.
  * apply eq_ax.
     unfold NKA_p;left;left;right;destruct E;simpl;auto 10.
  * assert (E':(eq_nom_untyped_p) e f)
      by (unfold rel,and_rel;tauto).
     apply (eq_nom_untyped_p_apply_perm p) in E'.
     eapply stronger_ax_proof;[|apply E'].
     unfold NKA,NKA_p,sg,rel,and_rel,stronger,or_rel;intuition.
  * apply eq_ax.
     left;right;destruct E;simpl;auto 10.
     rewrite (fresh_rm_perm p) in H;auto.
     rewrite (fresh_rm_perm p) in H;simpl in *;auto;tauto.
     rewrite (fresh_rm_perm p) in H;simpl in *;auto;tauto.
Qed.
Lemma NKA_θtp_NKA_tp_rm_perm_apply_perm :
  forall p : perm,
    Proper (NKA_θtp ==> eq NKA_tp)
           (rm_perm ∘ apply_perm p).
Proof.
  unfold Basics.compose;
  intros p e f [[[[[[[E|E]|E]|E]|E]|E]|E]|E].
  * destruct E;simpl;repeat eql;apply eq_ax;
    [apply kat_t1|apply kat_1t|apply kat_s1|apply kat_s2];
    apply has_type_rm_perm_apply_perm,has_type_apply_perm;
    now simpl_perm.
  * now apply (stronger_ax_proof (FFalse_spec _)),eq_perm_apply_perm.
  * now apply (stronger_ax_proof (FFalse_spec _)),eq_perm_ω_apply_perm.
  * now apply (stronger_ax_proof (FFalse_spec _)),
     eq_perm_elim_apply_perm.
  * eql;eql;eql;eqr;apply eq_ax;destruct E;simpl;auto 10.
  * eql;eql;eqr;apply eq_ax;destruct E;simpl;auto 10.
    now apply eq_ω_id;simpl_in;simpl_perm.
  * now eqr; apply eq_nom_typed_p_apply_perm. 
  * eql;eqr;apply eq_ax;destruct E;simpl;auto 10.
    -- eapply eqt_νω1; simpl in *;repeat simpl_beq;
       destruct H as (h1&_&h2).
       repeat split;try left;simpl_perm;eauto.
       2:apply has_type_rm_perm_apply_perm,has_type_apply_perm.
       2:rewrite mact_rm;simpl_perm;eauto.
       now simpl_in;simpl_perm.
    -- now apply eqt_νω2;rewrite perm_bij.
Qed.

(** ** Technical lemmas. *)
Lemma support_or_diff_NKA_p :
  (forall e f, (Bs ut) e f \/ ~ untyped e \/ ~ untyped f
               \/ ~ (NKA_p ⊢ e≡f)).
Proof.
  unfold Bs,ut,rel;
  intros e f.
  case (decidable_untyped e);try tauto.
  case (decidable_untyped f);try tauto;intros u1 u2.
  case (equiv_lst_b_spec (support e) (support f));try tauto;intros b.
  repeat right;intro h;apply b.
  eapply eq_support;eauto.
  unfold NKA_p;split_proper;
  auto with propersupport.
Qed.
Lemma ty'_or_diff_NKA_tp :
  (forall e f, ty' e f \/
               ~ typed e \/ ~ typed f
               \/ ~ (NKA_tp ⊢ e≡f)).
Proof.
  unfold ty';intros e f.
  case (decidable_typed e);try tauto.
  case (decidable_typed f);try tauto;intros (a&t1) (a'&t2).
  case (equiv_lst_b_spec a a');try tauto;intros b.
  - rewrite <-b in *;firstorder.
  - repeat right;intro h.
    apply eq_type with (A:=a') in h.
    rewrite h in t2;pose proof (type_unicity t1 t2);tauto.
    intro A;unfold NKA_tp;
    repeat (rewrite proper_or_split;split);
    auto with propertype.
Qed.
Lemma support_or_diff_NKA_θp :
  (forall e f, (Bs ut) e f \/ ~ untyped e \/ ~ untyped f
               \/ ~ (NKA_θp ⊢ e≡f)).
Proof.
  unfold Bs,ut,rel;
  intros e f.
  case (decidable_untyped e);try tauto.
  case (decidable_untyped f);try tauto;intros u1 u2.
  case (equiv_lst_b_spec (support e) (support f));try tauto;intros b.
  repeat right;intro h;apply b.
  eapply eq_support;eauto.
  unfold NKA_θp;repeat (rewrite proper_or_split;split);
  auto with propersupport.
Qed.
Lemma ty'_or_diff_NKA_θtp :
  (forall e f, ty' e f \/ ~ typed e \/ ~ typed f
               \/ ~ (NKA_θtp ⊢ e≡f)).
Proof.
  unfold ty';intros e f.
  case (decidable_typed e);try tauto.
  case (decidable_typed f);try tauto;intros (a&t1) (a'&t2).
  case (equiv_lst_b_spec a a');try tauto;intros b.
  - rewrite <-b in *;firstorder.
  - repeat right;intro h.
    apply eq_type with (A:=a') in h.
    rewrite h in t2;pose proof (type_unicity t1 t2);tauto.
    intro A;unfold NKA_θtp;repeat (rewrite proper_or_split;split);
    auto with propertype.
Qed.



(** * Good theories. *)

Definition good (th : TyEq) :=
  forall e f, (snd th) e f ->
              (fst th) e <-> (fst th) f.
Lemma goodNKAmpt : good NKAmpt.
Proof.
  unfold good,NKAmpt; simpl;intros e f E.
  pose proof (NKA_θtp_str_positive E) as p.
  pose proof (NKA_θtp_multi E) as m.
  assert (t:forall A, A ⊨ e <-> A ⊨ f) by
      (now intros; apply (NKA_θtp_has_type)).
  unfold typed, and_pred; firstorder.
Qed.
Lemma goodNKAspt : good NKAspt.
Proof.
  unfold good,NKAspt; simpl;intros e f E.
  pose proof (NKA_θtp_str_positive E) as p.
  pose proof (NKA_θtp_single E) as s.
  assert (t:forall A, A ⊨ e <-> A ⊨ f) by
      (now intros; apply (NKA_θtp_has_type)).
  unfold typed, and_pred; firstorder.
Qed.

Lemma goodNKAmpu : good NKAmpu.
Proof.
  unfold good,NKAmpu; simpl;intros e f E.
  pose proof (NKA_θp_str_positive E) as p.
  pose proof (NKA_θp_multi E) as m.
  pose proof (NKA_θp_untyped E) as u.
  unfold untyped, and_pred;firstorder.
Qed.
Lemma goodNKAspu : good NKAspu.
Proof.
  unfold good,NKAspu; simpl;intros e f E.
  pose proof (NKA_θp_str_positive E) as p.
  pose proof (NKA_θp_single E) as s.
  pose proof (NKA_θp_untyped E) as u.
  unfold typed, and_pred;firstorder.
Qed.

Lemma goodNKAnmpt : good NKAnmpt.
Proof.
    unfold good,NKAmpt; simpl;intros e f E.
    pose proof (NKA_tp_str_positive E) as p.
    pose proof (NKA_tp_multi E) as m.
    assert (t:forall A, A ⊨ e <-> A ⊨ f) by
        (now intros; apply (NKA_tp_has_type)).
    pose proof (NKA_tp_Np E) as n.
    unfold typed, and_pred; firstorder.
Qed.
Lemma goodNKAnspt : good NKAnspt.
Proof.
  unfold good,NKAmpt; simpl;intros e f E.
  pose proof (NKA_tp_str_positive E) as p.
  pose proof (NKA_tp_single E) as s.
  assert (t:forall A, A ⊨ e <-> A ⊨ f) by
      (now intros; apply (NKA_tp_has_type)).
  pose proof (NKA_tp_Np E) as n.
  unfold typed, and_pred; firstorder.
Qed.

Lemma goodNKAnmpu : good NKAnmpu.
Proof.
  unfold good,NKAmpt; simpl;intros e f E.
  pose proof (NKA_p_str_positive E) as p.
  pose proof (NKA_p_multi E) as m.
  pose proof (NKA_p_untyped E) as u.
  pose proof (NKA_p_Np E) as n.
  unfold untyped, and_pred;firstorder.
Qed.
Lemma goodNKAnspu : good NKAnspu.
Proof.
  unfold good,NKAmpt; simpl;intros e f E.
  pose proof (NKA_p_str_positive E) as p.
  pose proof (NKA_p_single E) as s.
  pose proof (NKA_p_untyped E) as u.
  pose proof (NKA_p_Np E) as n.
  unfold typed, and_pred;firstorder.
Qed.
